# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import operation
from google.cloud.aiplatform_v1.types import reasoning_engine as gca_reasoning_engine
from google.protobuf import field_mask_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "CreateReasoningEngineRequest",
        "CreateReasoningEngineOperationMetadata",
        "GetReasoningEngineRequest",
        "UpdateReasoningEngineRequest",
        "UpdateReasoningEngineOperationMetadata",
        "ListReasoningEnginesRequest",
        "ListReasoningEnginesResponse",
        "DeleteReasoningEngineRequest",
    },
)


class CreateReasoningEngineRequest(proto.Message):
    r"""Request message for
    [ReasoningEngineService.CreateReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.CreateReasoningEngine].

    Attributes:
        parent (str):
            Required. The resource name of the Location to create the
            ReasoningEngine in. Format:
            ``projects/{project}/locations/{location}``
        reasoning_engine (google.cloud.aiplatform_v1.types.ReasoningEngine):
            Required. The ReasoningEngine to create.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    reasoning_engine: gca_reasoning_engine.ReasoningEngine = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gca_reasoning_engine.ReasoningEngine,
    )


class CreateReasoningEngineOperationMetadata(proto.Message):
    r"""Details of
    [ReasoningEngineService.CreateReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.CreateReasoningEngine]
    operation.

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1.types.GenericOperationMetadata):
            The common part of the operation metadata.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )


class GetReasoningEngineRequest(proto.Message):
    r"""Request message for
    [ReasoningEngineService.GetReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.GetReasoningEngine].

    Attributes:
        name (str):
            Required. The name of the ReasoningEngine resource. Format:
            ``projects/{project}/locations/{location}/reasoningEngines/{reasoning_engine}``
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class UpdateReasoningEngineRequest(proto.Message):
    r"""Request message for
    [ReasoningEngineService.UpdateReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.UpdateReasoningEngine].

    Attributes:
        reasoning_engine (google.cloud.aiplatform_v1.types.ReasoningEngine):
            Required. The ReasoningEngine which replaces
            the resource on the server.
        update_mask (google.protobuf.field_mask_pb2.FieldMask):
            Optional. Mask specifying which fields to
            update.
    """

    reasoning_engine: gca_reasoning_engine.ReasoningEngine = proto.Field(
        proto.MESSAGE,
        number=1,
        message=gca_reasoning_engine.ReasoningEngine,
    )
    update_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=2,
        message=field_mask_pb2.FieldMask,
    )


class UpdateReasoningEngineOperationMetadata(proto.Message):
    r"""Details of
    [ReasoningEngineService.UpdateReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.UpdateReasoningEngine]
    operation.

    Attributes:
        generic_metadata (google.cloud.aiplatform_v1.types.GenericOperationMetadata):
            The common part of the operation metadata.
    """

    generic_metadata: operation.GenericOperationMetadata = proto.Field(
        proto.MESSAGE,
        number=1,
        message=operation.GenericOperationMetadata,
    )


class ListReasoningEnginesRequest(proto.Message):
    r"""Request message for
    [ReasoningEngineService.ListReasoningEngines][google.cloud.aiplatform.v1.ReasoningEngineService.ListReasoningEngines].

    Attributes:
        parent (str):
            Required. The resource name of the Location to list the
            ReasoningEngines from. Format:
            ``projects/{project}/locations/{location}``
        filter (str):
            Optional. The standard list filter. More detail in
            `AIP-160 <https://google.aip.dev/160>`__.
        page_size (int):
            Optional. The standard list page size.
        page_token (str):
            Optional. The standard list page token.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    filter: str = proto.Field(
        proto.STRING,
        number=2,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=3,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=4,
    )


class ListReasoningEnginesResponse(proto.Message):
    r"""Response message for
    [ReasoningEngineService.ListReasoningEngines][google.cloud.aiplatform.v1.ReasoningEngineService.ListReasoningEngines]

    Attributes:
        reasoning_engines (MutableSequence[google.cloud.aiplatform_v1.types.ReasoningEngine]):
            List of ReasoningEngines in the requested
            page.
        next_page_token (str):
            A token to retrieve the next page of results. Pass to
            [ListReasoningEnginesRequest.page_token][google.cloud.aiplatform.v1.ListReasoningEnginesRequest.page_token]
            to obtain that page.
    """

    @property
    def raw_page(self):
        return self

    reasoning_engines: MutableSequence[gca_reasoning_engine.ReasoningEngine] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message=gca_reasoning_engine.ReasoningEngine,
        )
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


class DeleteReasoningEngineRequest(proto.Message):
    r"""Request message for
    [ReasoningEngineService.DeleteReasoningEngine][google.cloud.aiplatform.v1.ReasoningEngineService.DeleteReasoningEngine].

    Attributes:
        name (str):
            Required. The name of the ReasoningEngine resource to be
            deleted. Format:
            ``projects/{project}/locations/{location}/reasoningEngines/{reasoning_engine}``
        force (bool):
            Optional. If set to true, child resources of this reasoning
            engine will also be deleted. Otherwise, the request will
            fail with FAILED_PRECONDITION error when the reasoning
            engine has undeleted child resources.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    force: bool = proto.Field(
        proto.BOOL,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



"""OpenAPI core contrib starlette middlewares module"""

from typing import Type

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from openapi_core import OpenAPI
from openapi_core.contrib.starlette.handlers import (
    StarletteOpenAPIErrorsHandler,
)
from openapi_core.contrib.starlette.handlers import (
    StarletteOpenAPIValidRequestHandler,
)
from openapi_core.contrib.starlette.integrations import StarletteIntegration
from openapi_core.contrib.starlette.requests import StarletteOpenAPIRequest
from openapi_core.contrib.starlette.responses import StarletteOpenAPIResponse


class StarletteOpenAPIMiddleware(StarletteIntegration, BaseHTTPMiddleware):
    valid_request_handler_cls = StarletteOpenAPIValidRequestHandler
    errors_handler = StarletteOpenAPIErrorsHandler()

    def __init__(
        self,
        app: ASGIApp,
        openapi: OpenAPI,
        request_cls: Type[StarletteOpenAPIRequest] = StarletteOpenAPIRequest,
        response_cls: Type[
            StarletteOpenAPIResponse
        ] = StarletteOpenAPIResponse,
    ):
        super().__init__(openapi)
        self.request_cls = request_cls
        self.response_cls = response_cls
        BaseHTTPMiddleware.__init__(self, app)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        valid_request_handler = self.valid_request_handler_cls(
            request, call_next
        )
        response = await self.handle_request(
            request, valid_request_handler, self.errors_handler
        )
        return await self.handle_response(
            request, response, self.errors_handler
        )



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from functools import lru_cache
from typing import TypedDict, Union

import requests

_GCP_METADATA_URL = "http://metadata.google.internal/computeMetadata/v1/"
_INSTANCE = "instance"
_RECURSIVE_PARAMS = {"recursive": "true"}
_GCP_METADATA_URL_HEADER = {"Metadata-Flavor": "Google"}
# Use a shorter timeout for connection so we won't block much if it's unreachable
_TIMEOUT = (2, 5)

_logger = logging.getLogger(__name__)


class Project(TypedDict):
    projectId: str


Attributes = TypedDict(
    "Attributes", {"cluster-location": str, "cluster-name": str}, total=False
)


class Instance(TypedDict):
    attributes: Attributes
    # id can be an integer on GCE VMs or a string on other environments
    id: Union[int, str]
    machineType: str
    name: str
    region: str
    zone: str


class Metadata(TypedDict):
    instance: Instance
    project: Project


class MetadataAccessException(Exception):
    pass


@lru_cache(maxsize=None)
def get_metadata() -> Metadata:
    """Get all instance and project metadata from the metadata server

    Cached for the lifetime of the process.
    """
    try:
        res = requests.get(
            f"{_GCP_METADATA_URL}",
            params=_RECURSIVE_PARAMS,
            headers=_GCP_METADATA_URL_HEADER,
            timeout=_TIMEOUT,
        )
        res.raise_for_status()
        all_metadata = res.json()
    except requests.RequestException as err:
        raise MetadataAccessException() from err
    return all_metadata


@lru_cache(maxsize=None)
def is_available() -> bool:
    try:
        requests.get(
            f"{_GCP_METADATA_URL}{_INSTANCE}/",
            headers=_GCP_METADATA_URL_HEADER,
            timeout=_TIMEOUT,
        ).raise_for_status()
    except requests.RequestException:
        _logger.debug(
            "Failed to make request to metadata server, assuming it's not available",
            exc_info=True,
        )
        return False
    return True



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import Sequence, TYPE_CHECKING, cast
import atexit
import warnings
from collections import deque
import concurrent.futures

from google.cloud.bigtable.data.exceptions import MutationsExceptionGroup
from google.cloud.bigtable.data.exceptions import FailedMutationEntryError
from google.cloud.bigtable.data._helpers import _get_retryable_errors
from google.cloud.bigtable.data._helpers import _get_timeouts
from google.cloud.bigtable.data._helpers import TABLE_DEFAULT

from google.cloud.bigtable.data.mutations import (
    _MUTATE_ROWS_REQUEST_MUTATION_LIMIT,
)
from google.cloud.bigtable.data.mutations import Mutation

from google.cloud.bigtable.data._cross_sync import CrossSync

if TYPE_CHECKING:
    from google.cloud.bigtable.data.mutations import RowMutationEntry

    if CrossSync.is_async:
        from google.cloud.bigtable.data._async.client import (
            _DataApiTargetAsync as TargetType,
        )
    else:
        from google.cloud.bigtable.data._sync_autogen.client import _DataApiTarget as TargetType  # type: ignore

__CROSS_SYNC_OUTPUT__ = "google.cloud.bigtable.data._sync_autogen.mutations_batcher"

# used to make more readable default values
_MB_SIZE = 1024 * 1024


@CrossSync.convert_class(sync_name="_FlowControl", add_mapping_for_name="_FlowControl")
class _FlowControlAsync:
    """
    Manages flow control for batched mutations. Mutations are registered against
    the FlowControl object before being sent, which will block if size or count
    limits have reached capacity. As mutations completed, they are removed from
    the FlowControl object, which will notify any blocked requests that there
    is additional capacity.

    Flow limits are not hard limits. If a single mutation exceeds the configured
    limits, it will be allowed as a single batch when the capacity is available.

    Args:
        max_mutation_count: maximum number of mutations to send in a single rpc.
            This corresponds to individual mutations in a single RowMutationEntry.
        max_mutation_bytes: maximum number of bytes to send in a single rpc.
    Raises:
        ValueError: if max_mutation_count or max_mutation_bytes is less than 0
    """

    def __init__(
        self,
        max_mutation_count: int,
        max_mutation_bytes: int,
    ):
        self._max_mutation_count = max_mutation_count
        self._max_mutation_bytes = max_mutation_bytes
        if self._max_mutation_count < 1:
            raise ValueError("max_mutation_count must be greater than 0")
        if self._max_mutation_bytes < 1:
            raise ValueError("max_mutation_bytes must be greater than 0")
        self._capacity_condition = CrossSync.Condition()
        self._in_flight_mutation_count = 0
        self._in_flight_mutation_bytes = 0

    def _has_capacity(self, additional_count: int, additional_size: int) -> bool:
        """
        Checks if there is capacity to send a new entry with the given size and count

        FlowControl limits are not hard limits. If a single mutation exceeds
        the configured flow limits, it will be sent in a single batch when
        previous batches have completed.

        Args:
            additional_count: number of mutations in the pending entry
            additional_size: size of the pending entry
        Returns:
            bool: True if there is capacity to send the pending entry, False otherwise
        """
        # adjust limits to allow overly large mutations
        acceptable_size = max(self._max_mutation_bytes, additional_size)
        acceptable_count = max(self._max_mutation_count, additional_count)
        # check if we have capacity for new mutation
        new_size = self._in_flight_mutation_bytes + additional_size
        new_count = self._in_flight_mutation_count + additional_count
        return new_size <= acceptable_size and new_count <= acceptable_count

    @CrossSync.convert
    async def remove_from_flow(
        self, mutations: RowMutationEntry | list[RowMutationEntry]
    ) -> None:
        """
        Removes mutations from flow control. This method should be called once
        for each mutation that was sent to add_to_flow, after the corresponding
        operation is complete.

        Args:
            mutations: mutation or list of mutations to remove from flow control
        """
        if not isinstance(mutations, list):
            mutations = [mutations]
        total_count = sum(len(entry.mutations) for entry in mutations)
        total_size = sum(entry.size() for entry in mutations)
        self._in_flight_mutation_count -= total_count
        self._in_flight_mutation_bytes -= total_size
        # notify any blocked requests that there is additional capacity
        async with self._capacity_condition:
            self._capacity_condition.notify_all()

    @CrossSync.convert
    async def add_to_flow(self, mutations: RowMutationEntry | list[RowMutationEntry]):
        """
        Generator function that registers mutations with flow control. As mutations
        are accepted into the flow control, they are yielded back to the caller,
        to be sent in a batch. If the flow control is at capacity, the generator
        will block until there is capacity available.

        Args:
            mutations: list mutations to break up into batches
        Yields:
            list[RowMutationEntry]:
                list of mutations that have reserved space in the flow control.
                Each batch contains at least one mutation.
        """
        if not isinstance(mutations, list):
            mutations = [mutations]
        start_idx = 0
        end_idx = 0
        while end_idx < len(mutations):
            start_idx = end_idx
            batch_mutation_count = 0
            # fill up batch until we hit capacity
            async with self._capacity_condition:
                while end_idx < len(mutations):
                    next_entry = mutations[end_idx]
                    next_size = next_entry.size()
                    next_count = len(next_entry.mutations)
                    if (
                        self._has_capacity(next_count, next_size)
                        # make sure not to exceed per-request mutation count limits
                        and (batch_mutation_count + next_count)
                        <= _MUTATE_ROWS_REQUEST_MUTATION_LIMIT
                    ):
                        # room for new mutation; add to batch
                        end_idx += 1
                        batch_mutation_count += next_count
                        self._in_flight_mutation_bytes += next_size
                        self._in_flight_mutation_count += next_count
                    elif start_idx != end_idx:
                        # we have at least one mutation in the batch, so send it
                        break
                    else:
                        # batch is empty. Block until we have capacity
                        await self._capacity_condition.wait_for(
                            lambda: self._has_capacity(next_count, next_size)
                        )
            yield mutations[start_idx:end_idx]


@CrossSync.convert_class(sync_name="MutationsBatcher")
class MutationsBatcherAsync:
    """
    Allows users to send batches using context manager API.

    Runs mutate_row,  mutate_rows, and check_and_mutate_row internally, combining
    to use as few network requests as required

    Will automatically flush the batcher:
    - every flush_interval seconds
    - after queue size reaches flush_limit_mutation_count
    - after queue reaches flush_limit_bytes
    - when batcher is closed or destroyed

    Args:
        table: table or autrhorized_view used to preform rpc calls
        flush_interval: Automatically flush every flush_interval seconds.
            If None, no time-based flushing is performed.
        flush_limit_mutation_count: Flush immediately after flush_limit_mutation_count
            mutations are added across all entries. If None, this limit is ignored.
        flush_limit_bytes: Flush immediately after flush_limit_bytes bytes are added.
        flow_control_max_mutation_count: Maximum number of inflight mutations.
        flow_control_max_bytes: Maximum number of inflight bytes.
        batch_operation_timeout: timeout for each mutate_rows operation, in seconds.
            If TABLE_DEFAULT, defaults to the Table's default_mutate_rows_operation_timeout.
        batch_attempt_timeout: timeout for each individual request, in seconds.
            If TABLE_DEFAULT, defaults to the Table's default_mutate_rows_attempt_timeout.
            If None, defaults to batch_operation_timeout.
        batch_retryable_errors: a list of errors that will be retried if encountered.
            Defaults to the Table's default_mutate_rows_retryable_errors.
    """

    def __init__(
        self,
        table: TargetType,
        *,
        flush_interval: float | None = 5,
        flush_limit_mutation_count: int | None = 1000,
        flush_limit_bytes: int = 20 * _MB_SIZE,
        flow_control_max_mutation_count: int = 100_000,
        flow_control_max_bytes: int = 100 * _MB_SIZE,
        batch_operation_timeout: float | TABLE_DEFAULT = TABLE_DEFAULT.MUTATE_ROWS,
        batch_attempt_timeout: float | None | TABLE_DEFAULT = TABLE_DEFAULT.MUTATE_ROWS,
        batch_retryable_errors: Sequence[type[Exception]]
        | TABLE_DEFAULT = TABLE_DEFAULT.MUTATE_ROWS,
    ):
        self._operation_timeout, self._attempt_timeout = _get_timeouts(
            batch_operation_timeout, batch_attempt_timeout, table
        )
        self._retryable_errors: list[type[Exception]] = _get_retryable_errors(
            batch_retryable_errors, table
        )

        self._closed = CrossSync.Event()
        self._target = table
        self._staged_entries: list[RowMutationEntry] = []
        self._staged_count, self._staged_bytes = 0, 0
        self._flow_control = CrossSync._FlowControl(
            flow_control_max_mutation_count, flow_control_max_bytes
        )
        self._flush_limit_bytes = flush_limit_bytes
        self._flush_limit_count = (
            flush_limit_mutation_count
            if flush_limit_mutation_count is not None
            else float("inf")
        )
        # used by sync class to run mutate_rows operations
        self._sync_rpc_executor = (
            concurrent.futures.ThreadPoolExecutor(max_workers=8)
            if not CrossSync.is_async
            else None
        )
        # used by sync class to manage flush_internal tasks
        self._sync_flush_executor = (
            concurrent.futures.ThreadPoolExecutor(max_workers=4)
            if not CrossSync.is_async
            else None
        )
        self._flush_timer = CrossSync.create_task(
            self._timer_routine, flush_interval, sync_executor=self._sync_flush_executor
        )
        self._flush_jobs: set[CrossSync.Future[None]] = set()
        # MutationExceptionGroup reports number of successful entries along with failures
        self._entries_processed_since_last_raise: int = 0
        self._exceptions_since_last_raise: int = 0
        # keep track of the first and last _exception_list_limit exceptions
        self._exception_list_limit: int = 10
        self._oldest_exceptions: list[Exception] = []
        self._newest_exceptions: deque[Exception] = deque(
            maxlen=self._exception_list_limit
        )
        # clean up on program exit
        atexit.register(self._on_exit)

    @CrossSync.convert
    async def _timer_routine(self, interval: float | None) -> None:
        """
        Set up a background task to flush the batcher every interval seconds

        If interval is None, an empty future is returned

        Args:
            flush_interval: Automatically flush every flush_interval seconds.
                If None, no time-based flushing is performed.
        """
        if not interval or interval <= 0:
            return None
        while not self._closed.is_set():
            # wait until interval has passed, or until closed
            await CrossSync.event_wait(
                self._closed, timeout=interval, async_break_early=False
            )
            if not self._closed.is_set() and self._staged_entries:
                self._schedule_flush()

    @CrossSync.convert
    async def append(self, mutation_entry: RowMutationEntry):
        """
        Add a new set of mutations to the internal queue

        Args:
            mutation_entry: new entry to add to flush queue
        Raises:
            RuntimeError: if batcher is closed
            ValueError: if an invalid mutation type is added
        """
        # TODO: return a future to track completion of this entry
        if self._closed.is_set():
            raise RuntimeError("Cannot append to closed MutationsBatcher")
        if isinstance(cast(Mutation, mutation_entry), Mutation):
            raise ValueError(
                f"invalid mutation type: {type(mutation_entry).__name__}. Only RowMutationEntry objects are supported by batcher"
            )
        self._staged_entries.append(mutation_entry)
        # start a new flush task if limits exceeded
        self._staged_count += len(mutation_entry.mutations)
        self._staged_bytes += mutation_entry.size()
        if (
            self._staged_count >= self._flush_limit_count
            or self._staged_bytes >= self._flush_limit_bytes
        ):
            self._schedule_flush()
            # yield to the event loop to allow flush to run
            await CrossSync.yield_to_event_loop()

    def _schedule_flush(self) -> CrossSync.Future[None] | None:
        """
        Update the flush task to include the latest staged entries

        Returns:
            Future[None] | None:
                future representing the background task, if started
        """
        if self._staged_entries:
            entries, self._staged_entries = self._staged_entries, []
            self._staged_count, self._staged_bytes = 0, 0
            new_task = CrossSync.create_task(
                self._flush_internal, entries, sync_executor=self._sync_flush_executor
            )
            if not new_task.done():
                self._flush_jobs.add(new_task)
                new_task.add_done_callback(self._flush_jobs.remove)
            return new_task
        return None

    @CrossSync.convert
    async def _flush_internal(self, new_entries: list[RowMutationEntry]):
        """
        Flushes a set of mutations to the server, and updates internal state

        Args:
            new_entries list of RowMutationEntry objects to flush
        """
        # flush new entries
        in_process_requests: list[CrossSync.Future[list[FailedMutationEntryError]]] = []
        async for batch in self._flow_control.add_to_flow(new_entries):
            batch_task = CrossSync.create_task(
                self._execute_mutate_rows, batch, sync_executor=self._sync_rpc_executor
            )
            in_process_requests.append(batch_task)
        # wait for all inflight requests to complete
        found_exceptions = await self._wait_for_batch_results(*in_process_requests)
        # update exception data to reflect any new errors
        self._entries_processed_since_last_raise += len(new_entries)
        self._add_exceptions(found_exceptions)

    @CrossSync.convert
    async def _execute_mutate_rows(
        self, batch: list[RowMutationEntry]
    ) -> list[FailedMutationEntryError]:
        """
        Helper to execute mutation operation on a batch

        Args:
            batch: list of RowMutationEntry objects to send to server
            timeout: timeout in seconds. Used as operation_timeout and attempt_timeout.
                If not given, will use table defaults
        Returns:
            list[FailedMutationEntryError]:
                list of FailedMutationEntryError objects for mutations that failed.
                FailedMutationEntryError objects will not contain index information
        """
        try:
            operation = CrossSync._MutateRowsOperation(
                self._target.client._gapic_client,
                self._target,
                batch,
                operation_timeout=self._operation_timeout,
                attempt_timeout=self._attempt_timeout,
                retryable_exceptions=self._retryable_errors,
            )
            await operation.start()
        except MutationsExceptionGroup as e:
            # strip index information from exceptions, since it is not useful in a batch context
            for subexc in e.exceptions:
                subexc.index = None
            return list(e.exceptions)
        finally:
            # mark batch as complete in flow control
            await self._flow_control.remove_from_flow(batch)
        return []

    def _add_exceptions(self, excs: list[Exception]):
        """
        Add new list of exceptions to internal store. To avoid unbounded memory,
        the batcher will store the first and last _exception_list_limit exceptions,
        and discard any in between.

        Args:
            excs: list of exceptions to add to the internal store
        """
        self._exceptions_since_last_raise += len(excs)
        if excs and len(self._oldest_exceptions) < self._exception_list_limit:
            # populate oldest_exceptions with found_exceptions
            addition_count = self._exception_list_limit - len(self._oldest_exceptions)
            self._oldest_exceptions.extend(excs[:addition_count])
            excs = excs[addition_count:]
        if excs:
            # populate newest_exceptions with remaining found_exceptions
            self._newest_exceptions.extend(excs[-self._exception_list_limit :])

    def _raise_exceptions(self):
        """
        Raise any unreported exceptions from background flush operations

        Raises:
            MutationsExceptionGroup: exception group with all unreported exceptions
        """
        if self._oldest_exceptions or self._newest_exceptions:
            oldest, self._oldest_exceptions = self._oldest_exceptions, []
            newest = list(self._newest_exceptions)
            self._newest_exceptions.clear()
            entry_count, self._entries_processed_since_last_raise = (
                self._entries_processed_since_last_raise,
                0,
            )
            exc_count, self._exceptions_since_last_raise = (
                self._exceptions_since_last_raise,
                0,
            )
            raise MutationsExceptionGroup.from_truncated_lists(
                first_list=oldest,
                last_list=newest,
                total_excs=exc_count,
                entry_count=entry_count,
            )

    @CrossSync.convert(sync_name="__enter__")
    async def __aenter__(self):
        """Allow use of context manager API"""
        return self

    @CrossSync.convert(sync_name="__exit__")
    async def __aexit__(self, exc_type, exc, tb):
        """
        Allow use of context manager API.

        Flushes the batcher and cleans up resources.
        """
        await self.close()

    @property
    def closed(self) -> bool:
        """
        Returns:
          - True if the batcher is closed, False otherwise
        """
        return self._closed.is_set()

    @CrossSync.convert
    async def close(self):
        """
        Flush queue and clean up resources
        """
        self._closed.set()
        self._flush_timer.cancel()
        self._schedule_flush()
        # shut down executors
        if self._sync_flush_executor:
            with self._sync_flush_executor:
                self._sync_flush_executor.shutdown(wait=True)
        if self._sync_rpc_executor:
            with self._sync_rpc_executor:
                self._sync_rpc_executor.shutdown(wait=True)
        await CrossSync.wait([*self._flush_jobs, self._flush_timer])
        atexit.unregister(self._on_exit)
        # raise unreported exceptions
        self._raise_exceptions()

    def _on_exit(self):
        """
        Called when program is exited. Raises warning if unflushed mutations remain
        """
        if not self._closed.is_set() and self._staged_entries:
            warnings.warn(
                f"MutationsBatcher for target {self._target!r} was not closed. "
                f"{len(self._staged_entries)} Unflushed mutations will not be sent to the server."
            )

    @staticmethod
    @CrossSync.convert
    async def _wait_for_batch_results(
        *tasks: CrossSync.Future[list[FailedMutationEntryError]]
        | CrossSync.Future[None],
    ) -> list[Exception]:
        """
        Takes in a list of futures representing _execute_mutate_rows tasks,
        waits for them to complete, and returns a list of errors encountered.

        Args:
            *tasks: futures representing _execute_mutate_rows or _flush_internal tasks
        Returns:
            list[Exception]:
                list of Exceptions encountered by any of the tasks. Errors are expected
                to be FailedMutationEntryError, representing a failed mutation operation.
                If a task fails with a different exception, it will be included in the
                output list. Successful tasks will not be represented in the output list.
        """
        if not tasks:
            return []
        exceptions: list[Exception] = []
        for task in tasks:
            if CrossSync.is_async:
                # futures don't need to be awaited in sync mode
                await task
            try:
                exc_list = task.result()
                if exc_list:
                    # expect a list of FailedMutationEntryError objects
                    for exc in exc_list:
                        # strip index information
                        exc.index = None
                    exceptions.extend(exc_list)
            except Exception as e:
                exceptions.append(e)
        return exceptions



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import feature_monitoring_stats
from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "Feature",
    },
)


class Feature(proto.Message):
    r"""Feature Metadata information.
    For example, color is a feature that describes an apple.

    Attributes:
        name (str):
            Immutable. Name of the Feature. Format:
            ``projects/{project}/locations/{location}/featurestores/{featurestore}/entityTypes/{entity_type}/features/{feature}``
            ``projects/{project}/locations/{location}/featureGroups/{feature_group}/features/{feature}``

            The last part feature is assigned by the client. The feature
            can be up to 64 characters long and can consist only of
            ASCII Latin letters A-Z and a-z, underscore(\_), and ASCII
            digits 0-9 starting with a letter. The value will be unique
            given an entity type.
        description (str):
            Description of the Feature.
        value_type (google.cloud.aiplatform_v1.types.Feature.ValueType):
            Immutable. Only applicable for Vertex AI
            Feature Store (Legacy). Type of Feature value.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Only applicable for Vertex AI
            Feature Store (Legacy). Timestamp when this
            EntityType was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Only applicable for Vertex AI
            Feature Store (Legacy). Timestamp when this
            EntityType was most recently updated.
        labels (MutableMapping[str, str]):
            Optional. The labels with user-defined
            metadata to organize your Features.
            Label keys and values can be no longer than 64
            characters (Unicode codepoints), can only
            contain lowercase letters, numeric characters,
            underscores and dashes. International characters
            are allowed.

            See https://goo.gl/xmQnxf for more information
            on and examples of labels. No more than 64 user
            labels can be associated with one Feature
            (System labels are excluded)."
            System reserved label keys are prefixed with
            "aiplatform.googleapis.com/" and are immutable.
        etag (str):
            Used to perform a consistent
            read-modify-write updates. If not set, a blind
            "overwrite" update happens.
        disable_monitoring (bool):
            Optional. Only applicable for Vertex AI Feature Store
            (Legacy). If not set, use the monitoring_config defined for
            the EntityType this Feature belongs to. Only Features with
            type
            ([Feature.ValueType][google.cloud.aiplatform.v1.Feature.ValueType])
            BOOL, STRING, DOUBLE or INT64 can enable monitoring.

            If set to true, all types of data monitoring are disabled
            despite the config on EntityType.
        monitoring_stats_anomalies (MutableSequence[google.cloud.aiplatform_v1.types.Feature.MonitoringStatsAnomaly]):
            Output only. Only applicable for Vertex AI
            Feature Store (Legacy). The list of historical
            stats and anomalies with specified objectives.
        version_column_name (str):
            Only applicable for Vertex AI Feature Store. The name of the
            BigQuery Table/View column hosting data for this version. If
            no value is provided, will use feature_id.
        point_of_contact (str):
            Entity responsible for maintaining this
            feature. Can be comma separated list of email
            addresses or URIs.
    """

    class ValueType(proto.Enum):
        r"""Only applicable for Vertex AI Legacy Feature Store.
        An enum representing the value type of a feature.

        Values:
            VALUE_TYPE_UNSPECIFIED (0):
                The value type is unspecified.
            BOOL (1):
                Used for Feature that is a boolean.
            BOOL_ARRAY (2):
                Used for Feature that is a list of boolean.
            DOUBLE (3):
                Used for Feature that is double.
            DOUBLE_ARRAY (4):
                Used for Feature that is a list of double.
            INT64 (9):
                Used for Feature that is INT64.
            INT64_ARRAY (10):
                Used for Feature that is a list of INT64.
            STRING (11):
                Used for Feature that is string.
            STRING_ARRAY (12):
                Used for Feature that is a list of String.
            BYTES (13):
                Used for Feature that is bytes.
            STRUCT (14):
                Used for Feature that is struct.
        """

        VALUE_TYPE_UNSPECIFIED = 0
        BOOL = 1
        BOOL_ARRAY = 2
        DOUBLE = 3
        DOUBLE_ARRAY = 4
        INT64 = 9
        INT64_ARRAY = 10
        STRING = 11
        STRING_ARRAY = 12
        BYTES = 13
        STRUCT = 14

    class MonitoringStatsAnomaly(proto.Message):
        r"""A list of historical
        [SnapshotAnalysis][google.cloud.aiplatform.v1.FeaturestoreMonitoringConfig.SnapshotAnalysis]
        or
        [ImportFeaturesAnalysis][google.cloud.aiplatform.v1.FeaturestoreMonitoringConfig.ImportFeaturesAnalysis]
        stats requested by user, sorted by
        [FeatureStatsAnomaly.start_time][google.cloud.aiplatform.v1.FeatureStatsAnomaly.start_time]
        descending.

        Attributes:
            objective (google.cloud.aiplatform_v1.types.Feature.MonitoringStatsAnomaly.Objective):
                Output only. The objective for each stats.
            feature_stats_anomaly (google.cloud.aiplatform_v1.types.FeatureStatsAnomaly):
                Output only. The stats and anomalies
                generated at specific timestamp.
        """

        class Objective(proto.Enum):
            r"""If the objective in the request is both
            Import Feature Analysis and Snapshot Analysis, this objective
            could be one of them. Otherwise, this objective should be the
            same as the objective in the request.

            Values:
                OBJECTIVE_UNSPECIFIED (0):
                    If it's OBJECTIVE_UNSPECIFIED, monitoring_stats will be
                    empty.
                IMPORT_FEATURE_ANALYSIS (1):
                    Stats are generated by Import Feature
                    Analysis.
                SNAPSHOT_ANALYSIS (2):
                    Stats are generated by Snapshot Analysis.
            """

            OBJECTIVE_UNSPECIFIED = 0
            IMPORT_FEATURE_ANALYSIS = 1
            SNAPSHOT_ANALYSIS = 2

        objective: "Feature.MonitoringStatsAnomaly.Objective" = proto.Field(
            proto.ENUM,
            number=1,
            enum="Feature.MonitoringStatsAnomaly.Objective",
        )
        feature_stats_anomaly: feature_monitoring_stats.FeatureStatsAnomaly = (
            proto.Field(
                proto.MESSAGE,
                number=2,
                message=feature_monitoring_stats.FeatureStatsAnomaly,
            )
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    description: str = proto.Field(
        proto.STRING,
        number=2,
    )
    value_type: ValueType = proto.Field(
        proto.ENUM,
        number=3,
        enum=ValueType,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=4,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=5,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=6,
    )
    etag: str = proto.Field(
        proto.STRING,
        number=7,
    )
    disable_monitoring: bool = proto.Field(
        proto.BOOL,
        number=12,
    )
    monitoring_stats_anomalies: MutableSequence[MonitoringStatsAnomaly] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=11,
            message=MonitoringStatsAnomaly,
        )
    )
    version_column_name: str = proto.Field(
        proto.STRING,
        number=106,
    )
    point_of_contact: str = proto.Field(
        proto.STRING,
        number=107,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import google.auth
import google.auth.transport.requests
import logging
import ray
import re
from immutabledict import immutabledict

from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.utils import resource_manager_utils

SUPPORTED_RAY_VERSIONS = immutabledict(
    {"2.9": "2.9.3", "2.33": "2.33.0", "2.42": "2.42.0", "2.47": "2.47.1"}
)
SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS = immutabledict(
    {
        "3.10": ("2.9", "2.33", "2.42"),
        "3.11": ("2.42", "2.47"),
    }
)
_V2_4_WARNING_MESSAGE = (
    "After google-cloud-aiplatform>1.53.0, using Ray version = 2.4 will result in "
    "an error. Please use Ray version = 2.33.0, 2.42.0 or 2.47.1 (default) instead."
)
_V2_9_WARNING_MESSAGE = (
    "In March 2025, using Ray version = 2.9 will result in an error. "
    "Please use Ray version = 2.33.0, 2.42.0 or 2.47.1 (default) instead."
)


# Artifact Repository available regions.
_AVAILABLE_REGIONS = ["us", "europe", "asia"]
# If region is not available, assume using the default region.
_DEFAULT_REGION = "us"

_PERSISTENT_RESOURCE_NAME_PATTERN = "projects/{}/locations/{}/persistentResources/{}"
_VALID_RESOURCE_NAME_REGEX = "[a-z][a-zA-Z0-9._-]{0,127}"
_DASHBOARD_URI_SUFFIX = "aiplatform-training.googleusercontent.com"


def valid_resource_name(resource_name):
    """Check if address is a valid resource name."""
    resource_name_split = resource_name.split("/")
    if not (
        len(resource_name_split) == 6
        and resource_name_split[0] == "projects"
        and resource_name_split[2] == "locations"
        and resource_name_split[4] == "persistentResources"
    ):
        raise ValueError(
            "[Ray on Vertex AI]: Address must be in the following "
            "format: vertex_ray://projects/<project_num>/locations/<region>/persistentResources/<pr_id> "
            "or vertex_ray://<pr_id>."
        )


def maybe_reconstruct_resource_name(address) -> str:
    """Reconstruct full persistent resource name if only id was given."""
    if re.match("^{}$".format(_VALID_RESOURCE_NAME_REGEX), address):
        # Assume only cluster name (persistent resource id) was given.
        logging.info(
            "[Ray on Vertex AI]: Cluster name was given as address, reconstructing full resource name"
        )
        return _PERSISTENT_RESOURCE_NAME_PATTERN.format(
            resource_manager_utils.get_project_number(
                initializer.global_config.project
            ),
            initializer.global_config.location,
            address,
        )

    return address


def get_local_ray_version():
    ray_version = ray.__version__.split(".")
    if len(ray_version) == 3:
        ray_version = ray_version[:2]
    return ".".join(ray_version)


def get_image_uri(ray_version, python_version, enable_cuda):
    """Image uri for a given ray version and python version."""
    if ray_version not in SUPPORTED_RAY_VERSIONS:
        raise ValueError(
            "[Ray on Vertex AI]: The supported Ray versions are %s (%s) and %s (%s)."
            % (
                list(SUPPORTED_RAY_VERSIONS.keys())[0],
                list(SUPPORTED_RAY_VERSIONS.values())[0],
                list(SUPPORTED_RAY_VERSIONS.keys())[1],
                list(SUPPORTED_RAY_VERSIONS.values())[1],
            )
        )
    if python_version is None:
        for pv, ray_versions in SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS.items():
            if ray_version in ray_versions:
                python_version = pv
                break
    if python_version not in SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS:
        raise ValueError(
            "[Ray on Vertex AI]: The supported Python versions are 3.10 or 3.11."
        )

    if ray_version not in SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS[python_version]:
        raise ValueError(
            "[Ray on Vertex AI]: The supported Ray version(s) for Python version %s: %s."
            % (
                python_version,
                SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS[python_version],
            )
        )

    location = initializer.global_config.location
    region = location.split("-")[0]
    if region not in _AVAILABLE_REGIONS:
        region = _DEFAULT_REGION
    ray_version = ray_version.replace(".", "-")
    python_version = python_version.replace(".", "")
    if enable_cuda:
        return f"{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.{ray_version}.py{python_version}:latest"
    else:
        return f"{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.{ray_version}.py{python_version}:latest"


def get_versions_from_image_uri(image_uri):
    """Get ray version and python version from image uri."""
    logging.info(f"[Ray on Vertex AI]: Getting versions from image uri: {image_uri}")
    image_label = image_uri.split("/")[-1].split(":")[0]
    py_version = image_label[-3] + "." + image_label[-2:]
    ray_version = image_label.split(".")[1].replace("-", ".")
    if (
        py_version in SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS
        and ray_version in SUPPORTED_RAY_VERSIONS_FROM_PYTHON_VERSIONS[py_version]
    ):
        return py_version, ray_version
    else:
        # May not parse custom image and get the versions correctly
        return None, None


def valid_dashboard_address(address):
    """Check if address is a valid dashboard uri."""
    return address.endswith(_DASHBOARD_URI_SUFFIX)


def get_bearer_token():
    """Get bearer token through Application Default Credentials."""
    creds, _ = google.auth.default(
        scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )

    # creds.valid is False, and creds.token is None
    # Need to refresh credentials to populate those
    auth_req = google.auth.transport.requests.Request()
    creds.refresh(auth_req)
    return creds.token



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
    Tuple,
    TYPE_CHECKING,
)
from google.api_core import retry as retries
from google.protobuf.message import Message
from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper

from google.cloud.bigtable.data.execute_query._byte_cursor import _ByteCursor
from google.cloud.bigtable.data._helpers import (
    _attempt_timeout_generator,
    _retry_exception_factory,
)
from google.cloud.bigtable.data.exceptions import (
    EarlyMetadataCallError,
    InvalidExecuteQueryResponse,
)
from google.cloud.bigtable.data.execute_query.values import QueryResultRow
from google.cloud.bigtable.data.execute_query.metadata import Metadata
from google.cloud.bigtable.data.execute_query._reader import (
    _QueryResultRowReader,
    _Reader,
)
from google.cloud.bigtable_v2.types.bigtable import (
    ExecuteQueryRequest as ExecuteQueryRequestPB,
    ExecuteQueryResponse,
)

from google.cloud.bigtable.data._cross_sync import CrossSync

if TYPE_CHECKING:
    if CrossSync.is_async:
        from google.cloud.bigtable.data import BigtableDataClientAsync as DataClientType
    else:
        from google.cloud.bigtable.data import BigtableDataClient as DataClientType

__CROSS_SYNC_OUTPUT__ = (
    "google.cloud.bigtable.data.execute_query._sync_autogen.execute_query_iterator"
)


def _has_resume_token(response: ExecuteQueryResponse) -> bool:
    response_pb = response._pb  # proto-plus attribute retrieval is slow.
    if response_pb.HasField("results"):
        results = response_pb.results
        return len(results.resume_token) > 0
    return False


@CrossSync.convert_class(sync_name="ExecuteQueryIterator")
class ExecuteQueryIteratorAsync:
    @CrossSync.convert(
        docstring_format_vars={
            "NO_LOOP": (
                "RuntimeError: if the instance is not created within an async event loop context.",
                "None",
            ),
            "TASK_OR_THREAD": ("asyncio Tasks", "threads"),
        }
    )
    def __init__(
        self,
        client: DataClientType,
        instance_id: str,
        app_profile_id: Optional[str],
        request_body: Dict[str, Any],
        prepare_metadata: Metadata,
        attempt_timeout: float | None,
        operation_timeout: float,
        req_metadata: Sequence[Tuple[str, str]] = (),
        retryable_excs: Sequence[type[Exception]] = (),
        column_info: dict[str, Message | EnumTypeWrapper] | None = None,
    ) -> None:
        """
        Collects responses from ExecuteQuery requests and parses them into QueryResultRows.

        **Please Note** this is not meant to be constructed directly by applications. It should always
        be created via the client. The constructor is subject to change.

        It is **not thread-safe**. It should not be used by multiple {TASK_OR_THREAD}.

        Args:
            client: bigtable client
            instance_id: id of the instance on which the query is executed
            request_body: dict representing the body of the ExecuteQueryRequest
            attempt_timeout: the time budget for an individual network request, in seconds.
                If it takes longer than this time to complete, the request will be cancelled with
                a DeadlineExceeded exception, and a retry will be attempted.
            operation_timeout: the time budget for the entire operation, in seconds.
                Failed requests will be retried within the budget
            req_metadata: metadata used while sending the gRPC request
            retryable_excs: a list of errors that will be retried if encountered.
            column_info: dict with mappings between column names and additional column information
                for protobuf deserialization.
        Raises:
            {NO_LOOP}
            :class:`ValueError <exceptions.ValueError>` as a safeguard if data is processed in an unexpected state
        """
        self._table_name = None
        self._app_profile_id = app_profile_id
        self._client = client
        self._instance_id = instance_id
        self._prepare_metadata: Metadata = prepare_metadata
        self._final_metadata: Metadata | None = None
        self._byte_cursor = _ByteCursor()
        self._reader: _Reader[QueryResultRow] = _QueryResultRowReader()
        self.has_received_token = False
        self._result_generator = self._next_impl()
        self._register_instance_task = None
        self._fully_consumed = False
        self._is_closed = False
        self._request_body = request_body
        self._attempt_timeout_gen = _attempt_timeout_generator(
            attempt_timeout, operation_timeout
        )
        self._stream = CrossSync.retry_target_stream(
            self._make_request_with_resume_token,
            retries.if_exception_type(*retryable_excs),
            retries.exponential_sleep_generator(0.01, 60, multiplier=2),
            operation_timeout,
            exception_factory=_retry_exception_factory,
        )
        self._req_metadata = req_metadata
        self._column_info = column_info
        try:
            self._register_instance_task = CrossSync.create_task(
                self._client._register_instance,
                self._instance_id,
                self.app_profile_id,
                id(self),
                sync_executor=self._client._executor,
            )
        except RuntimeError as e:
            raise RuntimeError(
                f"{self.__class__.__name__} must be created within an async event loop context."
            ) from e

    @property
    def is_closed(self) -> bool:
        """Returns True if the iterator is closed, False otherwise."""
        return self._is_closed

    @property
    def app_profile_id(self) -> Optional[str]:
        """Returns the app_profile_id of the iterator."""
        return self._app_profile_id

    @property
    def table_name(self) -> Optional[str]:
        """Returns the table_name of the iterator."""
        return self._table_name

    @CrossSync.convert
    async def _make_request_with_resume_token(self):
        """
        perfoms the rpc call using the correct resume token.
        """
        resume_token = self._byte_cursor.prepare_for_new_request()
        request = ExecuteQueryRequestPB(
            {
                **self._request_body,
                "resume_token": resume_token,
            }
        )
        return await self._client._gapic_client.execute_query(
            request,
            timeout=next(self._attempt_timeout_gen),
            metadata=self._req_metadata,
            retry=None,
        )

    @CrossSync.convert
    async def _next_impl(self) -> CrossSync.Iterator[QueryResultRow]:
        """
        Generator wrapping the response stream which parses the stream results
        and returns full `QueryResultRow`s.
        """
        try:
            async for response in self._stream:
                try:
                    # we've received a resume token, so we can finalize the metadata
                    if self._final_metadata is None and _has_resume_token(response):
                        self._finalize_metadata()

                    batches_to_parse = self._byte_cursor.consume(response)
                    if not batches_to_parse:
                        continue
                    # metadata must be set at this point since there must be a resume_token
                    # for byte_cursor to yield data
                    if not self.metadata:
                        raise ValueError(
                            "Error parsing response before finalizing metadata"
                        )
                    results = self._reader.consume(
                        batches_to_parse, self.metadata, self._column_info
                    )
                    if results is None:
                        continue

                except ValueError as e:
                    raise InvalidExecuteQueryResponse(
                        "Invalid ExecuteQuery response received"
                    ) from e

                for result in results:
                    yield result
            # this means the stream has finished with no responses. In that case we know the
            # latest_prepare_reponses was used successfully so we can finalize the metadata
            if self._final_metadata is None:
                self._finalize_metadata()
            self._fully_consumed = True
        finally:
            self._close_internal()

    @CrossSync.convert(sync_name="__next__", replace_symbols={"__anext__": "__next__"})
    async def __anext__(self) -> QueryResultRow:
        """
        Yields QueryResultRows representing the results of the query.

        :raises: :class:`ValueError <exceptions.ValueError>` as a safeguard if data is processed in an unexpected state
        """
        if self._is_closed:
            raise CrossSync.StopIteration
        return await self._result_generator.__anext__()

    @CrossSync.convert(sync_name="__iter__")
    def __aiter__(self):
        return self

    @CrossSync.convert
    def _finalize_metadata(self) -> None:
        """
        Sets _final_metadata to the metadata of the latest prepare_response.
        The iterator should call this after either the first resume token is received or the
        stream completes succesfully with no responses.

        This can't be set on init because the metadata will be able to change due to plan refresh.
        Plan refresh isn't implemented yet, but we want functionality to stay the same when it is.

        For example the following scenario for query "SELECT * FROM table":
          - Make a request, table has one column family 'cf'
          - Return an incomplete batch
          - request fails with transient error
          - Meanwhile the table has had a second column family added 'cf2'
          - Retry the request, get an error indicating the `prepared_query` has expired
          - Refresh the prepared_query and retry the request, the new prepared_query
            contains both 'cf' & 'cf2'
          - It sends a new incomplete batch and resets the old outdated batch
          - It send the next chunk with a checksum and resume_token, closing the batch.
        In this we need to use the updated schema from the refreshed prepare request.
        """
        self._final_metadata = self._prepare_metadata

    @property
    def metadata(self) -> Metadata:
        """
        Returns query metadata from the server or None if the iterator has been closed
        or if metadata has not been set yet.

        Metadata will not be set until the first row has been yielded or response with no rows
        completes.

        raises: :class:`EarlyMetadataCallError` when called before the first row has been returned
        or the iterator has completed with no rows in the response.
        """
        if not self._final_metadata:
            raise EarlyMetadataCallError()
        return self._final_metadata

    @CrossSync.convert
    async def close(self) -> None:
        """
        Cancel all background tasks. Should be called after all rows were processed.

        Called automatically by iterator

        :raises: :class:`ValueError <exceptions.ValueError>` if called in an invalid state
        """
        # this doesn't need to be async anymore but we wrap the sync api to avoid a breaking
        # change
        self._close_internal()

    def _close_internal(self) -> None:
        if self._is_closed:
            return
        # Throw an error if the iterator has been successfully consumed but there is
        # still buffered data
        if self._fully_consumed and not self._byte_cursor.empty():
            raise ValueError("Unexpected buffered data at end of executeQuery reqest")
        self._is_closed = True
        if self._register_instance_task is not None:
            self._register_instance_task.cancel()
        self._client._remove_instance_registration(
            self._instance_id, self.app_profile_id, id(self)
        )



# -*- coding: utf-8 -*-

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from google.auth import credentials as auth_credentials

from google.cloud.aiplatform.schedules import _Schedule as _ScheduleGa


class _Schedule(
    _ScheduleGa,
):
    """Preview Schedule resource for Vertex AI."""

    def __init__(
        self,
        credentials: auth_credentials.Credentials,
        project: str,
        location: str,
    ):
        """Retrieves a Schedule resource and instantiates its representation.
        Args:
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to create this Schedule.
                Overrides credentials set in aiplatform.init.
            project (str):
                Optional. The project that you want to run this Schedule in.
                If not set, the project set in aiplatform.init will be used.
            location (str):
                Optional. Location to create Schedule. If not set,
                location set in aiplatform.init will be used.
        """
        super().__init__(project=project, location=location, credentials=credentials)

    @property
    def cron_expression(self) -> str:
        """Current Schedule cron expression.

        Returns:
            Schedule cron expression.
        """
        return super().cron



# -*- coding: utf-8 -*-

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Launches Tensorboard Uploader for SDK."""

import threading
from typing import FrozenSet, Optional

from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer
from google.cloud.aiplatform.metadata.metadata import _experiment_tracker
from google.cloud.aiplatform.compat.services import (
    tensorboard_service_client,
)
from google.cloud.aiplatform.utils import TensorboardClientWithOverride

_LOGGER = base.Logger(__name__)
TensorboardServiceClient = tensorboard_service_client.TensorboardServiceClient


class _TensorBoardTracker:
    "Tracks TensorBoard uploader and uploader thread"

    def __init__(self):
        self._tensorboard_uploader = None

    def upload_tb_log(
        self,
        tensorboard_experiment_name: str,
        logdir: str,
        tensorboard_id: Optional[str] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        experiment_display_name: Optional[str] = None,
        run_name_prefix: Optional[str] = None,
        description: Optional[str] = None,
        verbosity: Optional[int] = 1,
        allowed_plugins: Optional[FrozenSet[str]] = None,
    ):
        """upload only the existing data in the logdir and then return immediately

        ```py
        Sample usage:
        aiplatform.init(location='us-central1', project='my-project')
        aiplatform.upload_tb_log(tensorboard_id='123',tensorboard_experiment_name='my-experiment',logdir='my-logdir')
        ```

        Args:
          tensorboard_experiment_name (str): Required. Name of this tensorboard
            experiment. Unique to the given
            projects/{project}/locations/{location}/tensorboards/{tensorboard_id}
          logdir (str): Required. The location of the TensorBoard logs that resides either in the local file system or Cloud Storage
          tensorboard_id (str): Optional. TensorBoard ID. If not set, tensorboard_id in aiplatform.init will be used.
          project (str): Optional. Project the TensorBoard is in. If not set, project set in aiplatform.init will be used.
          location (str): Optional. Location the TensorBoard is in. If not set, location set in aiplatform.init will be used.
          experiment_display_name (str): Optional. The display name of the
            experiment.
          run_name_prefix (str): Optional. If present, all runs created by this
            invocation will have their name prefixed by this value.
          description (str): Optional. String description to assign to the
            experiment.
          verbosity (str): Optional. Level of verbosity, an integer. Supported
            value: 0 - No upload statistics is printed. 1 - Print upload statistics
              while uploading data (default).
          allowed_plugins (FrozenSet[str]): Optional. List of additional allowed plugin names.
        """
        self._create_uploader(
            tensorboard_id=tensorboard_id,
            tensorboard_experiment_name=tensorboard_experiment_name,
            logdir=logdir,
            project=project,
            location=location,
            one_shot=True,
            experiment_display_name=experiment_display_name,
            run_name_prefix=run_name_prefix,
            description=description,
            verbosity=verbosity,
            allowed_plugins=allowed_plugins,
        ).start_uploading()
        _LOGGER.info("One time TensorBoard log upload completed.")

    def start_upload_tb_log(
        self,
        tensorboard_experiment_name: str,
        logdir: str,
        tensorboard_id: Optional[str] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        experiment_display_name: Optional[str] = None,
        run_name_prefix: Optional[str] = None,
        description: Optional[str] = None,
        allowed_plugins: Optional[FrozenSet[str]] = None,
    ):
        """Continues to listen for new data in the logdir and uploads when it appears.

        Note that after calling `start_upload_tb_log()` your thread will kept alive even if
        an exception is thrown. To ensure the thread gets shut down, put any code after
        `start_upload_tb_log()` and before `end_upload_tb_log()` in a `try` statement, and call
        `end_upload_tb_log()` in `finally`.

        ```py
        Sample usage:
        aiplatform.init(location='us-central1', project='my-project')
        aiplatform.start_upload_tb_log(tensorboard_id='123',tensorboard_experiment_name='my-experiment',logdir='my-logdir')

        try:
          # your code here
        finally:
          aiplatform.end_upload_tb_log()
        ```

        Args:
          tensorboard_experiment_name (str): Required. Name of this tensorboard
            experiment. Unique to the given
            projects/{project}/locations/{location}/tensorboards/{tensorboard_id}.
          logdir (str): Required. path of the log directory to upload
          tensorboard_id (str): Optional. TensorBoard ID. If not set, tensorboard_id in aiplatform.init will be used.
          project (str): Optional. Project the TensorBoard is in. If not set, project set in aiplatform.init will be used.
          location (str): Optional. Location the TensorBoard is in. If not set, location set in aiplatform.init will be used.
          experiment_display_name (str): Optional. The display name of the
            experiment.
          run_name_prefix (str): Optional. If present, all runs created by this
            invocation will have their name prefixed by this value.
          description (str): Optional. String description to assign to the
            experiment.
          allowed_plugins (FrozenSet[str]): Optional. List of additional allowed plugin names.
        """
        if self._tensorboard_uploader:
            _LOGGER.info(
                "Continuous upload running for {}. To start a new"
                " uploading process, please call end_upload_tb_log() to end the current one".format(
                    self._tensorboard_uploader.get_experiment_resource_name()
                )
            )
            return
        self._tensorboard_uploader = self._create_uploader(
            tensorboard_id=tensorboard_id,
            tensorboard_experiment_name=tensorboard_experiment_name,
            logdir=logdir,
            project=project,
            location=location,
            one_shot=False,
            experiment_display_name=experiment_display_name,
            run_name_prefix=run_name_prefix,
            description=description,
            verbosity=0,
            allowed_plugins=allowed_plugins,
        )
        threading.Thread(target=self._tensorboard_uploader.start_uploading).start()

    def end_upload_tb_log(self):
        """Ends the current TensorBoard uploader

        ```py
        aiplatform.start_upload_tb_log(...)
        ...
        aiplatform.end_upload_tb_log()
        ```
        """
        if not self._tensorboard_uploader:
            _LOGGER.info(
                "No uploader is running. To start a new uploader, call"
                " start_upload_tb_log()."
            )
            return
        self._tensorboard_uploader._end_uploading()
        self._tensorboard_uploader = None

    def _create_uploader(
        self,
        tensorboard_id: str,
        tensorboard_experiment_name: str,
        logdir: str,
        one_shot: bool,
        project: Optional[str] = None,
        location: Optional[str] = None,
        experiment_display_name: Optional[str] = None,
        run_name_prefix: Optional[str] = None,
        description: Optional[str] = None,
        verbosity: Optional[int] = 1,
        allowed_plugins: Optional[FrozenSet[str]] = None,
    ) -> "TensorBoardUploader":  # noqa: F821
        """Create a TensorBoardUploader and a TensorBoard Experiment

        Args:
          tensorboard_id (str): Required. TensorBoard ID.
          tensorboard_experiment_name (str): Required. Name of this tensorboard experiment. Unique to the given projects/{project}/locations/{location}/tensorboards/{tensorboard_id}
          logdir (str): Required. path of the log directory to upload
          one_shot (bool): Required. True: upload only the existing data in the logdir and then return immediately. False: continue to listen for new data in the logdir and upload them when it appears.
          project (str): Optional. Project the TensorBoard is in. If not set, project set in aiplatform.init will be used.
          location (str): Optional. Location the TensorBoard is in. If not set, location set in aiplatform.init will be used.
          experiment_display_name (str): Optional. The display name of the experiment.
          run_name_prefix (str): Optional. If present, all runs created by this invocation will have their name prefixed by this value.
          description (str): Optional. String description to assign to the experiment.
          verbosity (int)): Optional. Level of verbosity. Supported value: 0 - No upload statistics is printed. 1 - Print upload statistics while uploading data (default).
          allowed_plugins (FrozenSet[str]): Optional. List of additional allowed plugin names.

        Returns:
            An instance of TensorBoardUploader.

        Raise:
            ImportError if TensorBoard package is not installed.
            ValueError if TensorBoard ID is not specified in argument nor set with an Experiment
        """
        try:
            from google.cloud.aiplatform.tensorboard.uploader import (
                TensorBoardUploader,
            )
            from google.cloud.aiplatform.tensorboard import uploader_utils
            from google.cloud.aiplatform.tensorboard import (
                uploader_constants,
            )
        except ImportError:
            raise ImportError(
                "TensorBoard is not installed. Please install TensorBoard using pip install google-cloud-aiplatform[tensorboard] to use it in the Vertex SDK."
            )
        project = project or initializer.global_config.project
        location = location or initializer.global_config.location
        if tensorboard_id:
            tensorboard_resource_name = TensorboardServiceClient.tensorboard_path(
                project, location, tensorboard_id
            )
        else:
            if _experiment_tracker._get_global_tensorboard():
                tensorboard_resource_name = (
                    _experiment_tracker._get_global_tensorboard().resource_name
                )
            elif _experiment_tracker._experiment:
                if _experiment_tracker._experiment._lookup_backing_tensorboard():
                    tensorboard_resource_name = (
                        _experiment_tracker._experiment._lookup_backing_tensorboard().resource_name
                    )
                else:
                    raise ValueError(
                        f"No TensorBoard associated with experiment {initializer.global_config.experiment_name}. Please provide tensorboard_id in the argument."
                    )
            else:
                raise ValueError(
                    "No TensorBoard found. Please provide tensorboard_id in the argument."
                )

        api_client = initializer.global_config.create_client(
            client_class=TensorboardClientWithOverride,
            location_override=location,
        )
        (
            blob_storage_bucket,
            blob_storage_folder,
        ) = uploader_utils.get_blob_storage_bucket_and_folder(
            api_client, tensorboard_resource_name, project
        )

        plugins = (
            uploader_constants.ALLOWED_PLUGINS.union(allowed_plugins)
            if allowed_plugins
            else uploader_constants.ALLOWED_PLUGINS
        )

        tensorboard_uploader = TensorBoardUploader(
            experiment_name=tensorboard_experiment_name,
            tensorboard_resource_name=tensorboard_resource_name,
            experiment_display_name=experiment_display_name,
            blob_storage_bucket=blob_storage_bucket,
            blob_storage_folder=blob_storage_folder,
            allowed_plugins=plugins,
            writer_client=api_client,
            logdir=logdir,
            one_shot=one_shot,
            run_name_prefix=run_name_prefix,
            description=description,
            verbosity=verbosity,
        )
        tensorboard_uploader.create_experiment()
        print(
            "View your Tensorboard at https://{}.{}/experiment/{}".format(
                location,
                "tensorboard.googleusercontent.com",
                tensorboard_uploader.get_experiment_resource_name().replace("/", "+"),
            )
        )
        return tensorboard_uploader


_tensorboard_tracker = _TensorBoardTracker()



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import cached_content as gca_cached_content
from google.protobuf import field_mask_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "CreateCachedContentRequest",
        "GetCachedContentRequest",
        "UpdateCachedContentRequest",
        "DeleteCachedContentRequest",
        "ListCachedContentsRequest",
        "ListCachedContentsResponse",
    },
)


class CreateCachedContentRequest(proto.Message):
    r"""Request message for
    [GenAiCacheService.CreateCachedContent][google.cloud.aiplatform.v1.GenAiCacheService.CreateCachedContent].

    Attributes:
        parent (str):
            Required. The parent resource where the
            cached content will be created
        cached_content (google.cloud.aiplatform_v1.types.CachedContent):
            Required. The cached content to create
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    cached_content: gca_cached_content.CachedContent = proto.Field(
        proto.MESSAGE,
        number=2,
        message=gca_cached_content.CachedContent,
    )


class GetCachedContentRequest(proto.Message):
    r"""Request message for
    [GenAiCacheService.GetCachedContent][google.cloud.aiplatform.v1.GenAiCacheService.GetCachedContent].

    Attributes:
        name (str):
            Required. The resource name referring to the
            cached content
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class UpdateCachedContentRequest(proto.Message):
    r"""Request message for
    [GenAiCacheService.UpdateCachedContent][google.cloud.aiplatform.v1.GenAiCacheService.UpdateCachedContent].
    Only expire_time or ttl can be updated.

    Attributes:
        cached_content (google.cloud.aiplatform_v1.types.CachedContent):
            Required. The cached content to update
        update_mask (google.protobuf.field_mask_pb2.FieldMask):
            Required. The list of fields to update.
    """

    cached_content: gca_cached_content.CachedContent = proto.Field(
        proto.MESSAGE,
        number=1,
        message=gca_cached_content.CachedContent,
    )
    update_mask: field_mask_pb2.FieldMask = proto.Field(
        proto.MESSAGE,
        number=2,
        message=field_mask_pb2.FieldMask,
    )


class DeleteCachedContentRequest(proto.Message):
    r"""Request message for
    [GenAiCacheService.DeleteCachedContent][google.cloud.aiplatform.v1.GenAiCacheService.DeleteCachedContent].

    Attributes:
        name (str):
            Required. The resource name referring to the
            cached content
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )


class ListCachedContentsRequest(proto.Message):
    r"""Request to list CachedContents.

    Attributes:
        parent (str):
            Required. The parent, which owns this
            collection of cached contents.
        page_size (int):
            Optional. The maximum number of cached
            contents to return. The service may return fewer
            than this value. If unspecified, some default
            (under maximum) number of items will be
            returned. The maximum value is 1000; values
            above 1000 will be coerced to 1000.
        page_token (str):
            Optional. A page token, received from a previous
            ``ListCachedContents`` call. Provide this to retrieve the
            subsequent page.

            When paginating, all other parameters provided to
            ``ListCachedContents`` must match the call that provided the
            page token.
    """

    parent: str = proto.Field(
        proto.STRING,
        number=1,
    )
    page_size: int = proto.Field(
        proto.INT32,
        number=2,
    )
    page_token: str = proto.Field(
        proto.STRING,
        number=3,
    )


class ListCachedContentsResponse(proto.Message):
    r"""Response with a list of CachedContents.

    Attributes:
        cached_contents (MutableSequence[google.cloud.aiplatform_v1.types.CachedContent]):
            List of cached contents.
        next_page_token (str):
            A token, which can be sent as ``page_token`` to retrieve the
            next page. If this field is omitted, there are no subsequent
            pages.
    """

    @property
    def raw_page(self):
        return self

    cached_contents: MutableSequence[gca_cached_content.CachedContent] = (
        proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message=gca_cached_content.CachedContent,
        )
    )
    next_page_token: str = proto.Field(
        proto.STRING,
        number=2,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import base64
import datetime
import json
import logging
import re
from base64 import b64encode
from functools import partial
from typing import (
    Any,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    TextIO,
    cast,
)

import google.auth
from google.api.monitored_resource_pb2 import (  # pylint: disable = no-name-in-module
    MonitoredResource,
)
from google.cloud.logging_v2.services.logging_service_v2 import (
    LoggingServiceV2Client,
)
from google.cloud.logging_v2.services.logging_service_v2.transports.grpc import (
    LoggingServiceV2GrpcTransport,
)
from google.cloud.logging_v2.types.log_entry import LogEntry
from google.cloud.logging_v2.types.logging import WriteLogEntriesRequest
from google.logging.type.log_severity_pb2 import (  # pylint: disable = no-name-in-module
    LogSeverity,
)
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import (  # pylint: disable = no-name-in-module
    Struct,
)
from google.protobuf.timestamp_pb2 import (  # pylint: disable = no-name-in-module
    Timestamp,
)
from opentelemetry.exporter.cloud_logging.version import __version__
from opentelemetry.resourcedetector.gcp_resource_detector._mapping import (
    get_monitored_resource,
)
from opentelemetry.sdk import version as opentelemetry_sdk_version
from opentelemetry.sdk._logs import LogData
from opentelemetry.sdk._logs.export import LogExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.trace import format_span_id, format_trace_id
from opentelemetry.util.types import AnyValue
from proto.datetime_helpers import (  # type: ignore[import]
    DatetimeWithNanoseconds,
)

DEFAULT_MAX_ENTRY_SIZE = 256000  # 256 KB
DEFAULT_MAX_REQUEST_SIZE = 10000000  # 10 MB

HTTP_REQUEST_ATTRIBUTE_KEY = "gcp.http_request"
LOG_NAME_ATTRIBUTE_KEY = "gcp.log_name"
SOURCE_LOCATION_ATTRIBUTE_KEY = "gcp.source_location"
TRACE_SAMPLED_ATTRIBUTE_KEY = "gcp.trace_sampled"
PROJECT_ID_ATTRIBUTE_KEY = "gcp.project_id"
_OTEL_SDK_VERSION = opentelemetry_sdk_version.__version__
_USER_AGENT = f"opentelemetry-python {_OTEL_SDK_VERSION}; google-cloud-logging-exporter {__version__}"

# Set user-agent metadata, see https://github.com/grpc/grpc/issues/23644 and default options
# from
# https://github.com/googleapis/python-logging/blob/5309478c054d0f2b9301817fd835f2098f51dc3a/google/cloud/logging_v2/services/logging_service_v2/transports/grpc.py#L179-L182
_OPTIONS = [
    ("grpc.max_send_message_length", -1),
    ("grpc.max_receive_message_length", -1),
    ("grpc.primary_user_agent", _USER_AGENT),
]

# severityMapping maps the integer severity level values from OTel [0-24]
# to matching Cloud Logging severity levels.
SEVERITY_MAPPING: dict[int, int] = {
    0: LogSeverity.DEFAULT,
    1: LogSeverity.DEBUG,
    2: LogSeverity.DEBUG,
    3: LogSeverity.DEBUG,
    4: LogSeverity.DEBUG,
    5: LogSeverity.DEBUG,
    6: LogSeverity.DEBUG,
    7: LogSeverity.DEBUG,
    8: LogSeverity.DEBUG,
    9: LogSeverity.INFO,
    10: LogSeverity.INFO,
    11: LogSeverity.NOTICE,
    12: LogSeverity.NOTICE,
    13: LogSeverity.WARNING,
    14: LogSeverity.WARNING,
    15: LogSeverity.WARNING,
    16: LogSeverity.WARNING,
    17: LogSeverity.ERROR,
    18: LogSeverity.ERROR,
    19: LogSeverity.ERROR,
    20: LogSeverity.ERROR,
    21: LogSeverity.CRITICAL,
    22: LogSeverity.CRITICAL,
    23: LogSeverity.ALERT,
    24: LogSeverity.EMERGENCY,
}

INVALID_LOG_NAME_MESSAGE = "%s is not a valid log name. log name must be <512 characters and only contain characters: A-Za-z0-9/-_."


class _GenAiJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, bytes):
            return b64encode(o).decode()
        return super().default(o)


def _convert_any_value_to_string(value: Any) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, bytes):
        return base64.b64encode(value).decode()
    if isinstance(value, (int, float, str)):
        return str(value)
    if isinstance(value, (list, tuple, Mapping)):
        return json.dumps(value, separators=(",", ":"), cls=_GenAiJsonEncoder)
    try:
        return str(value)
    except Exception as exc:  # pylint: disable=broad-except
        logging.exception(
            "Error mapping AnyValue to string, this field will not be added to the LogEntry: %s",
            exc,
        )
        return ""


# Be careful not to mutate original body. Make copies of anything that needs to change.
def _sanitized_body(
    body: Mapping[str, AnyValue],
) -> MutableMapping[str, AnyValue]:
    new_body: MutableMapping[str, AnyValue] = {}
    for key, value in body.items():
        if (
            isinstance(value, Sequence)
            and len(value) > 0
            and isinstance(value[0], bytes)
        ):
            # Should not be possible for a non-bytes value to be present. AnyValue requires Sequence be of one type, and above
            # we verified the first value is type bytes.
            new_body[key] = [
                base64.b64encode(v).decode() if isinstance(v, bytes) else v
                for v in value
            ]
        elif (
            isinstance(value, Sequence)
            and len(value) > 0
            and isinstance(value[0], Mapping)
        ):
            # Should not be possible for a non-mapping value to be present. AnyValue requires Sequence be of one type, and above
            # we verified the first value is type mapping.
            new_body[key] = [
                _sanitized_body(x) if isinstance(x, Mapping) else x
                for x in value
            ]
        elif isinstance(value, bytes):
            new_body[key] = base64.b64encode(value).decode()
        elif isinstance(value, Mapping):
            new_body[key] = _sanitized_body(value)
        else:
            new_body[key] = value
    return new_body


def _set_payload_in_log_entry(log_entry: LogEntry, body: AnyValue):
    struct = Struct()
    if isinstance(body, Mapping):
        sanitized = _sanitized_body(body)
        try:
            struct.update(sanitized)
            log_entry.json_payload = struct
        except Exception as exc:  # pylint: disable=broad-except
            logging.exception(
                "Error mapping LogRecord.body to Struct, will write log with empty body: %s",
                exc,
            )
    elif isinstance(body, bytes):
        json_str = body.decode("utf-8", errors="replace")
        json_dict = json.loads(json_str)
        if isinstance(json_dict, Mapping):
            struct.update(json_dict)
            log_entry.json_payload = struct
        else:
            log_entry.text_payload = base64.b64encode(body).decode()
    elif body is not None:
        log_entry.text_payload = _convert_any_value_to_string(body)


def is_log_id_valid(log_id: str) -> bool:
    return len(log_id) < 512 and not bool(
        re.search(r"[^A-Za-z0-9\-_/\.]", log_id)
    )


def _get_monitored_resource(
    resource: Optional[Resource],
) -> MonitoredResource | None:
    if not resource:
        return None

    # TODO: Remove temporary special case for Vertex Agent Engine
    # https://github.com/GoogleCloudPlatform/opentelemetry-operations-python/issues/444
    cloud_resource_id = resource.attributes.get("cloud.resource_id")
    if isinstance(cloud_resource_id, str) and (
        match := re.match(
            r"//aiplatform\.googleapis\.com/projects/(?P<project_id>[^/]+)"
            r"/locations/(?P<location>[^/]+)"
            r"/reasoningEngines/(?P<agent_engine_id>[^/]+)",
            cloud_resource_id,
        )
    ):
        location = match.group("location")
        agent_engine_id = match.group("agent_engine_id")
        # https://cloud.google.com/monitoring/api/resources#tag_aiplatform.googleapis.com/ReasoningEngine
        return MonitoredResource(
            type="aiplatform.googleapis.com/ReasoningEngine",
            labels={
                # Intentionally omit the project ID
                "location": location,
                "reasoning_engine_id": agent_engine_id,
            },
        )

    monitored_resource_data = get_monitored_resource(resource)
    if not monitored_resource_data:
        return None

    return MonitoredResource(
        type=monitored_resource_data.type,
        labels=monitored_resource_data.labels,
    )


class CloudLoggingExporter(LogExporter):
    def __init__(
        self,
        project_id: Optional[str] = None,
        default_log_name: Optional[str] = None,
        client: Optional[LoggingServiceV2Client] = None,
        *,
        structured_json_file: Optional[TextIO] = None,
    ) -> None:
        """Create a CloudLoggingExporter

        Args:
            project_id: The GCP project ID to which the logs will be sent. If not
                provided, the exporter will infer it from Application Default Credentials.
            default_log_name: The default log name to use for log entries.
                If not provided, a default name will be used.
            client: An optional `LoggingServiceV2Client` instance to use for
                sending logs. If not provided and ``structured_json_file`` is not provided, a
                new client will be created. Passing both ``client`` and
                ``structured_json_file`` is not supported.
            structured_json_file: An optional file-like object (like `sys.stdout`) to write
                logs to in Cloud Logging `structured JSON format
                <https://cloud.google.com/logging/docs/structured-logging>`_. If provided,
                ``client`` must not be provided and logs will only be written to the file-like
                object.
        """

        self.project_id: str
        if not project_id:
            _, default_project_id = google.auth.default()
            self.project_id = str(default_project_id)
        else:
            self.project_id = project_id

        if default_log_name:
            self.default_log_name = default_log_name
        else:
            self.default_log_name = "otel_python_inprocess_log_name_temp"

        if client and structured_json_file:
            raise ValueError(
                "Cannot specify both client and structured_json_file"
            )

        if structured_json_file:
            self._write_log_entries = partial(
                self._write_log_entries_to_file, structured_json_file
            )
        else:
            client = client or LoggingServiceV2Client(
                transport=LoggingServiceV2GrpcTransport(
                    channel=LoggingServiceV2GrpcTransport.create_channel(
                        options=_OPTIONS,
                    )
                )
            )
            self._write_log_entries = partial(
                self._write_log_entries_to_client, client
            )

    def pick_log_id(self, log_name_attr: Any, event_name: str | None) -> str:
        if log_name_attr and isinstance(log_name_attr, str):
            if is_log_id_valid(log_name_attr):
                return log_name_attr.replace("/", "%2F")
            logging.warning(INVALID_LOG_NAME_MESSAGE, log_name_attr)
        if event_name and is_log_id_valid(event_name):
            return event_name.replace("/", "%2F")
        return self.default_log_name

    def export(self, batch: Sequence[LogData]):
        now = datetime.datetime.now()
        log_entries = []
        for log_data in batch:
            log_entry = LogEntry()
            log_record = log_data.log_record
            attributes = log_record.attributes or {}
            project_id = str(
                attributes.get(PROJECT_ID_ATTRIBUTE_KEY, self.project_id)
            )
            log_entry.log_name = f"projects/{project_id}/logs/{self.pick_log_id(attributes.get(LOG_NAME_ATTRIBUTE_KEY), log_record.event_name)}"
            # If timestamp is unset fall back to observed_time_unix_nano as recommended,
            # see https://github.com/open-telemetry/opentelemetry-proto/blob/4abbb78/opentelemetry/proto/logs/v1/logs.proto#L176-L179
            ts = Timestamp()
            if log_record.timestamp or log_record.observed_timestamp:
                ts.FromNanoseconds(
                    log_record.timestamp or log_record.observed_timestamp
                )
            else:
                ts.FromDatetime(now)
            log_entry.timestamp = ts
            if monitored_resource := _get_monitored_resource(
                log_record.resource
            ):
                log_entry.resource = monitored_resource
            log_entry.trace_sampled = (
                log_record.trace_flags is not None
                and log_record.trace_flags.sampled
            )
            if log_record.trace_id:
                log_entry.trace = f"projects/{project_id}/traces/{format_trace_id(log_record.trace_id)}"
            if log_record.span_id:
                log_entry.span_id = format_span_id(log_record.span_id)
            if (
                log_record.severity_number
                and log_record.severity_number.value in SEVERITY_MAPPING
            ):
                log_entry.severity = SEVERITY_MAPPING[  # type: ignore[assignment]
                    log_record.severity_number.value  # type: ignore[index]
                ]
            log_entry.labels = {
                k: _convert_any_value_to_string(v)
                for k, v in attributes.items()
            }
            if log_record.event_name:
                log_entry.labels["event.name"] = log_record.event_name
            _set_payload_in_log_entry(log_entry, log_record.body)
            log_entries.append(log_entry)

        self._write_log_entries(log_entries)

    @staticmethod
    def _write_log_entries_to_file(file: TextIO, log_entries: list[LogEntry]):
        """Formats logs into the Cloud Logging structured log format, and writes them to the
        specified file-like object

        See https://cloud.google.com/logging/docs/structured-logging
        """
        # TODO: this is not resilient to exceptions which can cause recursion when using OTel's
        # logging handler. See
        # https://github.com/open-telemetry/opentelemetry-python/issues/4261 for outstanding
        # issue in OTel.

        for entry in log_entries:
            json_dict: dict[str, Any] = {}

            # These are not added in export() so not added to the JSON here.
            # - httpRequest
            # - logging.googleapis.com/sourceLocation
            # - logging.googleapis.com/operation
            # - logging.googleapis.com/insertId

            # https://cloud.google.com/logging/docs/agent/logging/configuration#timestamp-processing
            timestamp = cast(DatetimeWithNanoseconds, entry.timestamp)
            json_dict["time"] = timestamp.rfc3339()

            json_dict["severity"] = LogSeverity.Name(
                cast(LogSeverity.ValueType, entry.severity)
            )
            json_dict["logging.googleapis.com/labels"] = dict(entry.labels)
            json_dict["logging.googleapis.com/spanId"] = entry.span_id
            json_dict[
                "logging.googleapis.com/trace_sampled"
            ] = entry.trace_sampled
            json_dict["logging.googleapis.com/trace"] = entry.trace

            if entry.text_payload:
                json_dict["message"] = entry.text_payload
            if entry.json_payload:
                json_dict.update(
                    MessageToDict(LogEntry.pb(entry).json_payload)
                )

            # Use dumps to avoid invalid json written to the stream if serialization fails for any reason
            file.write(
                json.dumps(json_dict, separators=(",", ":"), sort_keys=True)
                + "\n"
            )

    @staticmethod
    def _write_log_entries_to_client(
        client: LoggingServiceV2Client, log_entries: list[LogEntry]
    ):
        batch: list[LogEntry] = []
        batch_byte_size = 0
        for entry in log_entries:
            msg_size = LogEntry.pb(entry).ByteSize()
            if msg_size > DEFAULT_MAX_ENTRY_SIZE:
                logging.warning(
                    "Cannot write log that is %s bytes which exceeds Cloud Logging's maximum limit of %s bytes.",
                    msg_size,
                    DEFAULT_MAX_ENTRY_SIZE,
                )
                continue
            if msg_size + batch_byte_size > DEFAULT_MAX_REQUEST_SIZE:
                try:
                    client.write_log_entries(
                        WriteLogEntriesRequest(
                            entries=batch, partial_success=True
                        )
                    )
                # pylint: disable=broad-except
                except Exception as ex:
                    logging.error(
                        "Error while writing to Cloud Logging", exc_info=ex
                    )
                batch = [entry]
                batch_byte_size = msg_size
            else:
                batch.append(entry)
                batch_byte_size += msg_size
        if batch:
            try:
                client.write_log_entries(
                    WriteLogEntriesRequest(entries=batch, partial_success=True)
                )
            # pylint: disable=broad-except
            except Exception as ex:
                logging.error(
                    "Error while writing to Cloud Logging", exc_info=ex
                )

    def shutdown(self):
        pass



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define Cloud Logging API Sinks."""

from google.cloud.exceptions import NotFound


class Sink(object):
    """Sinks represent filtered exports for log entries.

    See https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks
    """

    def __init__(
        self, name, *, filter_=None, parent=None, destination=None, client=None
    ):
        """
        Args:
            name (str): The name of the sink.
            parent(Optional[str]): The resource in which to create the sink:

                ::

                    "projects/[PROJECT_ID]"
                    "organizations/[ORGANIZATION_ID]"
                    "billingAccounts/[BILLING_ACCOUNT_ID]"
                    "folders/[FOLDER_ID]".

                Defaults to the project stored on the client.
            filter_ (Optional[str]): The advanced logs filter expression defining
                the entries exported by the sink.
            destination (Optional[str]): Destination URI for the entries exported by the sink.
                If not passed, the instance should already exist, to
                be refreshed via :meth:`reload`.
            client (Optional[~logging_v2.client.Client]): A client which holds
                credentials and project configuration for the sink (which requires a project).
        """
        self.name = name
        self.filter_ = filter_
        self.destination = destination
        self._client = client
        self._parent = parent
        self._writer_identity = None

    @property
    def client(self):
        """Client bound to the sink."""
        return self._client

    @property
    def parent(self):
        """Parent resource of the sink (project, organization, billingAccount, or folder)."""
        if self._parent is None:
            self._parent = f"projects/{self.client.project}"
        return self._parent

    @property
    def full_name(self):
        """Fully-qualified name used in sink APIs"""
        return f"{self.parent}/sinks/{self.name}"

    @property
    def path(self):
        """URL path for the sink's APIs"""
        return f"/{self.full_name}"

    @property
    def writer_identity(self):
        """Identity used for exports via the sink"""
        return self._writer_identity

    def _update_from_api_repr(self, resource):
        """Helper for API methods returning sink resources."""
        self.destination = resource["destination"]
        self.filter_ = resource.get("filter")
        self._writer_identity = resource.get("writerIdentity")

    @classmethod
    def from_api_repr(cls, resource, client, *, parent=None):
        """Construct a sink given its API representation

        Args:
            resource (dict): sink resource representation returned from the API
            client (~logging_v2.client.Client): Client which holds
                credentials and project configuration for the sink.
            parent(Optional[str]): The resource in which to create the sink:

                ::

                    "projects/[PROJECT_ID]"
                    "organizations/[ORGANIZATION_ID]"
                    "billingAccounts/[BILLING_ACCOUNT_ID]"
                    "folders/[FOLDER_ID]".

                Defaults to the project stored on the client.

        Returns:
            ~logging_v2.sink.Sink: Sink parsed from ``resource``.

        Raises:
            ValueError: if ``client`` is not ``None`` and the
                project from the resource does not agree with the project
                from the client.
        """
        sink_name = resource["name"]
        instance = cls(sink_name, client=client, parent=parent)
        instance._update_from_api_repr(resource)
        return instance

    def _require_client(self, client):
        """Check client or verify over-ride. Also sets ``parent``.

        Args:
            client (Union[None, ~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.

        Returns:
            ~logging_v2.client.Client: The client passed in
                or the currently bound client.
        """
        if client is None:
            client = self._client
        return client

    def create(self, *, client=None, unique_writer_identity=False):
        """Create the sink via a PUT request

        See
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks/create

        Args:
            client (Optional[~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.
            unique_writer_identity (Optional[bool]): Determines the kind of
                IAM identity returned as writer_identity in the new sink.
        """
        client = self._require_client(client)
        resource = client.sinks_api.sink_create(
            self.parent,
            self.name,
            self.filter_,
            self.destination,
            unique_writer_identity=unique_writer_identity,
        )
        self._update_from_api_repr(resource)

    def exists(self, *, client=None):
        """Test for the existence of the sink via a GET request

        See
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks/get

        Args:
            client (Optional[~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.

        Returns:
            bool: Boolean indicating existence of the sink.
        """
        client = self._require_client(client)

        try:
            client.sinks_api.sink_get(self.full_name)
        except NotFound:
            return False
        else:
            return True

    def reload(self, *, client=None):
        """Sync local sink configuration via a GET request

        See
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks/get

        Args:
            client (Optional[~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.
        """
        client = self._require_client(client)
        resource = client.sinks_api.sink_get(self.full_name)
        self._update_from_api_repr(resource)

    def update(self, *, client=None, unique_writer_identity=False):
        """Update sink configuration via a PUT request

        See
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks/update

        Args:
            client (Optional[~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.
            unique_writer_identity (Optional[bool]): Determines the kind of
                IAM identity returned as writer_identity in the new sink.
        """
        client = self._require_client(client)
        resource = client.sinks_api.sink_update(
            self.full_name,
            self.filter_,
            self.destination,
            unique_writer_identity=unique_writer_identity,
        )
        self._update_from_api_repr(resource)

    def delete(self, *, client=None):
        """Delete a sink via a DELETE request

        See
        https://cloud.google.com/logging/docs/reference/v2/rest/v2/projects.sinks/delete

        Args:
            client (Optional[~logging_v2.client.Client]):
                The client to use.  If not passed, falls back to the
                ``client`` stored on the current sink.
        """
        client = self._require_client(client)
        client.sinks_api.sink_delete(self.full_name)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Helper functions used in various places in the library.
"""
from __future__ import annotations

from typing import Sequence, List, Tuple, TYPE_CHECKING, Union
import time
import enum
from collections import namedtuple
from google.cloud.bigtable.data.read_rows_query import ReadRowsQuery

from google.api_core import exceptions as core_exceptions
from google.api_core.retry import RetryFailureReason
from google.cloud.bigtable.data.exceptions import RetryExceptionGroup

if TYPE_CHECKING:
    import grpc
    from google.cloud.bigtable.data._async.client import _DataApiTargetAsync
    from google.cloud.bigtable.data._sync_autogen.client import _DataApiTarget

"""
Helper functions used in various places in the library.
"""

# Type alias for the output of sample_keys
RowKeySamples = List[Tuple[bytes, int]]

# type alias for the output of query.shard()
ShardedQuery = List[ReadRowsQuery]

# used by read_rows_sharded to limit how many requests are attempted in parallel
_CONCURRENCY_LIMIT = 10

# used to identify an active bigtable resource that needs to be warmed through PingAndWarm
# each instance/app_profile_id pair needs to be individually tracked
_WarmedInstanceKey = namedtuple(
    "_WarmedInstanceKey", ["instance_name", "app_profile_id"]
)


# enum used on method calls when table defaults should be used
class TABLE_DEFAULT(enum.Enum):
    # default for mutate_row, sample_row_keys, check_and_mutate_row, and read_modify_write_row
    DEFAULT = "DEFAULT"
    # default for read_rows, read_rows_stream, read_rows_sharded, row_exists, and read_row
    READ_ROWS = "READ_ROWS_DEFAULT"
    # default for bulk_mutate_rows and mutations_batcher
    MUTATE_ROWS = "MUTATE_ROWS_DEFAULT"


def _attempt_timeout_generator(
    per_request_timeout: float | None, operation_timeout: float
):
    """
    Generator that yields the timeout value for each attempt of a retry loop.

    Will return per_request_timeout until the operation_timeout is approached,
    at which point it will return the remaining time in the operation_timeout.

    Args:
        per_request_timeout: The timeout value to use for each request, in seconds.
            If None, the operation_timeout will be used for each request.
        operation_timeout: The timeout value to use for the entire operationm in seconds.
    Yields:
        float: The timeout value to use for the next request, in seonds
    """
    per_request_timeout = (
        per_request_timeout if per_request_timeout is not None else operation_timeout
    )
    deadline = operation_timeout + time.monotonic()
    while True:
        yield max(0, min(per_request_timeout, deadline - time.monotonic()))


def _retry_exception_factory(
    exc_list: list[Exception],
    reason: RetryFailureReason,
    timeout_val: float | None,
) -> tuple[Exception, Exception | None]:
    """
    Build retry error based on exceptions encountered during operation

    Args:
        exc_list: list of exceptions encountered during operation
        is_timeout: whether the operation failed due to timeout
        timeout_val: the operation timeout value in seconds, for constructing
            the error message
    Returns:
        tuple[Exception, Exception|None]:
            tuple of the exception to raise, and a cause exception if applicable
    """
    if reason == RetryFailureReason.TIMEOUT:
        timeout_val_str = f"of {timeout_val:0.1f}s " if timeout_val is not None else ""
        # if failed due to timeout, raise deadline exceeded as primary exception
        source_exc: Exception = core_exceptions.DeadlineExceeded(
            f"operation_timeout{timeout_val_str} exceeded"
        )
    elif exc_list:
        # otherwise, raise non-retryable error as primary exception
        source_exc = exc_list.pop()
    else:
        source_exc = RuntimeError("failed with unspecified exception")
    # use the retry exception group as the cause of the exception
    cause_exc: Exception | None = RetryExceptionGroup(exc_list) if exc_list else None
    source_exc.__cause__ = cause_exc
    return source_exc, cause_exc


def _get_timeouts(
    operation: float | TABLE_DEFAULT,
    attempt: float | None | TABLE_DEFAULT,
    table: "_DataApiTargetAsync" | "_DataApiTarget",
) -> tuple[float, float]:
    """
    Convert passed in timeout values to floats, using table defaults if necessary.

    attempt will use operation value if None, or if larger than operation.

    Will call _validate_timeouts on the outputs, and raise ValueError if the
    resulting timeouts are invalid.

    Args:
        operation: The timeout value to use for the entire operation, in seconds.
        attempt: The timeout value to use for each attempt, in seconds.
        table: The table to use for default values.
    Returns:
        tuple[float, float]: A tuple of (operation_timeout, attempt_timeout)
    """
    # load table defaults if necessary
    if operation == TABLE_DEFAULT.DEFAULT:
        final_operation = table.default_operation_timeout
    elif operation == TABLE_DEFAULT.READ_ROWS:
        final_operation = table.default_read_rows_operation_timeout
    elif operation == TABLE_DEFAULT.MUTATE_ROWS:
        final_operation = table.default_mutate_rows_operation_timeout
    else:
        final_operation = operation
    if attempt == TABLE_DEFAULT.DEFAULT:
        attempt = table.default_attempt_timeout
    elif attempt == TABLE_DEFAULT.READ_ROWS:
        attempt = table.default_read_rows_attempt_timeout
    elif attempt == TABLE_DEFAULT.MUTATE_ROWS:
        attempt = table.default_mutate_rows_attempt_timeout

    return _align_timeouts(final_operation, attempt)


def _align_timeouts(operation: float, attempt: float | None) -> tuple[float, float]:
    """
    Convert passed in timeout values to floats.

    attempt will use operation value if None, or if larger than operation.

    Will call _validate_timeouts on the outputs, and raise ValueError if the
    resulting timeouts are invalid.

    Args:
        operation: The timeout value to use for the entire operation, in seconds.
        attempt: The timeout value to use for each attempt, in seconds.
    Returns:
        tuple[float, float]: A tuple of (operation_timeout, attempt_timeout)
    """
    if attempt is None:
        # no timeout specified, use operation timeout for both
        final_attempt = operation
    else:
        # cap attempt timeout at operation timeout
        final_attempt = min(attempt, operation) if operation else attempt

    _validate_timeouts(operation, final_attempt, allow_none=False)
    return operation, final_attempt


def _validate_timeouts(
    operation_timeout: float, attempt_timeout: float | None, allow_none: bool = False
):
    """
    Helper function that will verify that timeout values are valid, and raise
    an exception if they are not.

    Args:
        operation_timeout: The timeout value to use for the entire operation, in seconds.
        attempt_timeout: The timeout value to use for each attempt, in seconds.
        allow_none: If True, attempt_timeout can be None. If False, None values will raise an exception.
    Raises:
        ValueError: if operation_timeout or attempt_timeout are invalid.
    """
    if operation_timeout is None:
        raise ValueError("operation_timeout cannot be None")
    if operation_timeout <= 0:
        raise ValueError("operation_timeout must be greater than 0")
    if not allow_none and attempt_timeout is None:
        raise ValueError("attempt_timeout must not be None")
    elif attempt_timeout is not None:
        if attempt_timeout <= 0:
            raise ValueError("attempt_timeout must be greater than 0")


def _get_error_type(
    call_code: Union["grpc.StatusCode", int, type[Exception]]
) -> type[Exception]:
    """Helper function for ensuring the object is an exception type.
    If it is not, the proper GoogleAPICallError type is infered from the status
    code.

    Args:
      - call_code: Exception type or gRPC status code.
    """
    if isinstance(call_code, type):
        return call_code
    else:
        return type(core_exceptions.from_grpc_status(call_code, ""))


def _get_retryable_errors(
    call_codes: Sequence["grpc.StatusCode" | int | type[Exception]] | TABLE_DEFAULT,
    table: "_DataApiTargetAsync" | "_DataApiTarget",
) -> list[type[Exception]]:
    """
    Convert passed in retryable error codes to a list of exception types.

    Args:
        call_codes: The error codes to convert. Can be a list of grpc.StatusCode values,
            int values, or Exception types, or a TABLE_DEFAULT value.
        table: The table to use for default values.
    Returns:
        list[type[Exception]]: A list of exception types to retry on.
    """
    # load table defaults if necessary
    if call_codes == TABLE_DEFAULT.DEFAULT:
        call_codes = table.default_retryable_errors
    elif call_codes == TABLE_DEFAULT.READ_ROWS:
        call_codes = table.default_read_rows_retryable_errors
    elif call_codes == TABLE_DEFAULT.MUTATE_ROWS:
        call_codes = table.default_mutate_rows_retryable_errors

    return [_get_error_type(e) for e in call_codes]



# -*- coding: utf-8 -*-

# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from typing import List, Dict, Optional
from google.cloud.aiplatform import initializer
from google.auth import credentials as auth_credentials
from google.cloud.aiplatform import base
from google.cloud.aiplatform import utils
from google.cloud.aiplatform.compat.types import (
    feature_view_sync as gca_feature_view_sync,
    feature_view as gca_feature_view,
    feature_online_store_service as fos_service,
)
import vertexai.resources.preview.feature_store.utils as fs_utils

_LOGGER = base.Logger(__name__)


class FeatureView(base.VertexAiResourceNounWithFutureManager):
    """Class for managing Feature View resources."""

    client_class = utils.FeatureOnlineStoreAdminClientWithOverride

    _resource_noun = "featureViews"
    _getter_method = "get_feature_view"
    _list_method = "list_feature_views"
    _delete_method = "delete_feature_view"
    _parse_resource_name_method = "parse_feature_view_path"
    _format_resource_name_method = "feature_view_path"
    _gca_resource: gca_feature_view.FeatureView
    _online_store_client: utils.FeatureOnlineStoreClientWithOverride

    _online_store_clients_with_connection_options: Dict[
        fs_utils.ConnectionOptions, utils.FeatureOnlineStoreClientWithOverride
    ] = None

    def __init__(
        self,
        name: str,
        feature_online_store_id: Optional[str] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ):
        """Retrieves an existing managed feature view.

        Args:
            name:
                The resource name
                (`projects/.../locations/.../featureOnlineStores/.../featureViews/...`)
                or ID.
            feature_online_store_id:
                The feature online store ID. Must be passed in if name is an ID
                and not a resource path.
            project:
                Project to retrieve the feature view from. If unset, the project
                set in aiplatform.init will be used.
            location:
                Location to retrieve the feature view from. If not set, location
                set in aiplatform.init will be used.
            credentials:
                Custom credentials to use to retrieve this feature view.
                Overrides credentials set in aiplatform.init.
        """

        super().__init__(
            project=project,
            location=location,
            credentials=credentials,
            resource_name=name,
        )

        if re.fullmatch(
            r"projects/.+/locations/.+/featureOnlineStores/.+/featureViews/.+",
            name,
        ):
            feature_view = name
        else:
            from .feature_online_store import FeatureOnlineStore

            # Construct the feature view path using feature online store ID if
            # only the feature view ID is provided.
            if not feature_online_store_id:
                raise ValueError(
                    "Since feature view is not provided as a path, please specify"
                    + " feature_online_store_id."
                )

            feature_online_store_path = utils.full_resource_name(
                resource_name=feature_online_store_id,
                resource_noun=FeatureOnlineStore._resource_noun,
                parse_resource_name_method=FeatureOnlineStore._parse_resource_name,
                format_resource_name_method=FeatureOnlineStore._format_resource_name,
            )

            feature_view = f"{feature_online_store_path}/featureViews/{name}"

        self._gca_resource = self._get_gca_resource(resource_name=feature_view)

    def _get_online_store_client(
        self, connection_options: Optional[fs_utils.ConnectionOptions] = None
    ) -> utils.FeatureOnlineStoreClientWithOverride:
        """Return the online store client.

        Also sets the `_online_store_client` attr if not set yet. Note that if
        `connection_options` is passed in, the `_online_store_client` attr will
        not be set - only the client will be returned. If the same
        `connection_options` is passed in, this code will return the same
        (cached) client as previously built.
        """
        if getattr(self, "_online_store_client", None):
            return self._online_store_client

        fos_name = fs_utils.get_feature_online_store_name(self.resource_name)
        from .feature_online_store import FeatureOnlineStore

        fos = FeatureOnlineStore(name=fos_name)

        if connection_options:
            # Check if we have a previously client created for these
            # connection_options.
            if self._online_store_clients_with_connection_options is None:
                self._online_store_clients_with_connection_options = {}
            if connection_options in self._online_store_clients_with_connection_options:
                return self._online_store_clients_with_connection_options[
                    connection_options
                ]
            host = connection_options.host

            if isinstance(
                connection_options.transport,
                fs_utils.ConnectionOptions.InsecureGrpcChannel,
            ):
                import grpc
                from google.cloud.aiplatform_v1.services import (
                    feature_online_store_service as feature_online_store_service_v1,
                )
                from google.cloud.aiplatform_v1beta1.services import (
                    feature_online_store_service as feature_online_store_service_v1beta1,
                )

                gapic_client_class = (
                    utils.FeatureOnlineStoreClientWithOverride.get_gapic_client_class()
                )
                gapic_client_class_to_transport_class = {
                    feature_online_store_service_v1.client.FeatureOnlineStoreServiceClient: (
                        feature_online_store_service_v1.transports.grpc.FeatureOnlineStoreServiceGrpcTransport
                    ),
                    feature_online_store_service_v1beta1.client.FeatureOnlineStoreServiceClient: (
                        feature_online_store_service_v1beta1.transports.grpc.FeatureOnlineStoreServiceGrpcTransport
                    ),
                }
                if gapic_client_class not in gapic_client_class_to_transport_class:
                    raise ValueError(
                        f"Unexpected gapic class '{gapic_client_class}' used by internal client."
                    )

                transport_class = gapic_client_class_to_transport_class[
                    gapic_client_class
                ]

                client = gapic_client_class(
                    transport=transport_class(
                        channel=grpc.insecure_channel(host + ":10002")
                    ),
                )

                self._online_store_clients_with_connection_options[
                    connection_options
                ] = client
                return client
            else:
                raise ValueError(
                    f"Unsupported connection transport type, got transport: {connection_options.transport}"
                )

        if fos._gca_resource.bigtable.auto_scaling:
            # This is Bigtable online store.
            _LOGGER.info(f"Connecting to Bigtable online store name {fos_name}")
            self._online_store_client = initializer.global_config.create_client(
                client_class=utils.FeatureOnlineStoreClientWithOverride,
                credentials=self.credentials,
                location_override=self.location,
            )
            return self._online_store_client

        if (
            fos._gca_resource.dedicated_serving_endpoint.private_service_connect_config.enable_private_service_connect
        ):
            raise ValueError(
                "Use `connection_options` to specify an IP address. Required for optimized online store with private service connect."
            )

        # From here, optimized serving with public endpoint.
        if not fos._gca_resource.dedicated_serving_endpoint.public_endpoint_domain_name:
            raise fs_utils.PublicEndpointNotFoundError(
                "Public endpoint is not created yet for the optimized online store:"
                f"{fos_name}. Please run sync and wait for it to complete."
            )

        _LOGGER.info(
            f"Public endpoint for the optimized online store {fos_name} is"
            f" {fos._gca_resource.dedicated_serving_endpoint.public_endpoint_domain_name}"
        )
        self._online_store_client = initializer.global_config.create_client(
            client_class=utils.FeatureOnlineStoreClientWithOverride,
            credentials=self.credentials,
            location_override=self.location,
            prediction_client=True,
            api_path_override=fos._gca_resource.dedicated_serving_endpoint.public_endpoint_domain_name,
        )
        return self._online_store_client

    @classmethod
    def list(
        cls,
        feature_online_store_id: str,
        filter: Optional[str] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ) -> List["FeatureView"]:
        """List all feature view under feature_online_store_id.

        Example Usage:
        ```
        feature_views = vertexai.preview.FeatureView.list(
            feature_online_store_id="my_fos",
            filter=labels.label_key=label_value)
        ```
        Args:
            feature_online_store_id:
                Parentfeature online store ID.
            filter:
                Filter to apply on the returned feature online store.
            project:
                Project to use to get a list of feature views. If unset, the
                project set in aiplatform.init will be used.
            location:
                Location to use to get a list feature views. If not set,
                location set in aiplatform.init will be used.
            credentials:
                Custom credentials to use to get a list of feature views.
                Overrides credentials set in aiplatform.init.

        Returns:
            List[FeatureView] - list of FeatureView resource object.
        """
        from .feature_online_store import FeatureOnlineStore

        fos = FeatureOnlineStore(
            name=feature_online_store_id,
            project=project,
            location=location,
            credentials=credentials,
        )
        return cls._list(
            filter=filter, credentials=credentials, parent=fos.resource_name
        )

    @base.optional_sync()
    def delete(self, sync: bool = True) -> None:
        """Deletes this feature view.

        WARNING: This deletion is permanent.

        Args:
            sync:
                Whether to execute this deletion synchronously. If False, this
                method will be executed in concurrent Future and any downstream
                object will be immediately returned and synced when the Future
                has completed.
        """
        lro = getattr(self.api_client, self._delete_method)(name=self.resource_name)
        _LOGGER.log_delete_with_lro(self, lro)
        lro.result()
        _LOGGER.log_delete_complete(self)

    def sync(self) -> "FeatureViewSync":
        """Starts an on-demand Sync for the FeatureView.

        Args: None

        Returns:
            "FeatureViewSync" - FeatureViewSync instance
        """
        sync_method = getattr(self.api_client, self.FeatureViewSync.sync_method())

        sync_request = {
            "feature_view": self.resource_name,
        }
        sync_response = sync_method(request=sync_request)

        return self.FeatureViewSync(name=sync_response.feature_view_sync)

    def get_sync(self, name) -> "FeatureViewSync":
        """Gets the FeatureViewSync resource for the given name.

        Args:
            name: The resource ID

        Returns:
            "FeatureViewSync" - FeatureViewSync instance
        """
        feature_view_path = self.resource_name
        feature_view_sync = f"{feature_view_path}/featureViewSyncs/{name}"
        return self.FeatureViewSync(name=feature_view_sync)

    def list_syncs(
        self,
        filter: Optional[str] = None,
    ) -> List["FeatureViewSync"]:
        """List all feature view under this FeatureView.

        Args:
            parent_resource_name: Fully qualified name of the parent FeatureView
              resource.
            filter: Filter to apply on the returned feature online store.

        Returns:
            List[FeatureViewSync] - list of FeatureViewSync resource object.
        """

        return self.FeatureViewSync._list(
            filter=filter, credentials=self.credentials, parent=self.resource_name
        )

    def read(
        self,
        key: List[str],
        connection_options: Optional[fs_utils.ConnectionOptions] = None,
        request_timeout: Optional[float] = None,
    ) -> fs_utils.FeatureViewReadResponse:
        """Read the feature values from FeatureView.

          Example Usage:
            Read feature view. Use this for Bigtable online stores and for
            Optimized online stores that use public endpoint.
            ```
            data = vertexai.preview.FeatureView(
                name='feature_view_name', feature_online_store_id='fos_name')
                .read(key=[12345, 6789])
                .to_dict()
            ```

            Read feature view using IP with an insecure gRPC channel. Use this
            for optimized online stores using private service connect.
            ```
            data = vertexai.preview.FeatureView(
                name='feature_view_name', feature_online_store_id='fos_name')
                .read(
                    key=[12345, 6789],
                    connection_options=fs_utils.ConnectionOptions(
                        host="<ip>",
                        transport=fs_utils.ConnectionOptions.InsecureGrpcChannel()))
                .to_dict()
            ```
        Args:
            key: The request key to read feature values for.
            connection_options:
                If specified, use these options to connect to a host for sending
                requests instead of the default
                `<region>-aiplatform.googleapis.com` or the feature online
                store's public endpoint.

        Returns:
            "FeatureViewReadResponse" - FeatureViewReadResponse object. It is
            intermediate class that can be further converted by to_dict() or
            to_proto().
        """
        self.wait()

        online_store_client = self._get_online_store_client(
            connection_options=connection_options
        )

        response = online_store_client.fetch_feature_values(
            feature_view=self.resource_name,
            data_key=fos_service.FeatureViewDataKey(
                composite_key=fos_service.FeatureViewDataKey.CompositeKey(parts=key)
            ),
            timeout=request_timeout,
        )
        return fs_utils.FeatureViewReadResponse(response)

    def search(
        self,
        entity_id: Optional[str] = None,
        embedding_value: Optional[List[float]] = None,
        neighbor_count: Optional[int] = None,
        string_filters: Optional[
            List[fos_service.NearestNeighborQuery.StringFilter]
        ] = None,
        per_crowding_attribute_neighbor_count: Optional[int] = None,
        return_full_entity: bool = False,
        approximate_neighbor_candidates: Optional[int] = None,
        leaf_nodes_search_fraction: Optional[float] = None,
        request_timeout: Optional[float] = None,
    ) -> fs_utils.SearchNearestEntitiesResponse:
        """Search the nearest entities from FeatureView.

        Example Usage:
        ```
          data = vertexai.preview.FeatureView(
              name='feature_view_name', feature_online_store_id='fos_name')
            .search(entity_id='sample_entity')
            .to_dict()
        ```
        Args:
            entity_id: The entity id whose similar entities should be searched
              for.
            embedding_value: The embedding vector that be used for similar
              search.
            neighbor_count: The number of similar entities to be retrieved
              from feature view for each query.
            string_filters: The list of string filters.
            per_crowding_attribute_neighbor_count: Crowding is a constraint on a
            neighbor list produced by nearest neighbor search requiring that
              no more than sper_crowding_attribute_neighbor_count of the k
              neighbors returned have the same value of crowding_attribute.
              It's used for improving result diversity.
            return_full_entity: If true, return full entities including the
              features other than embeddings.
            approximate_neighbor_candidates: The number of neighbors to find via
              approximate search before exact reordering is performed; if set,
              this value must be > neighbor_count.
            leaf_nodes_search_fraction: The fraction of the number of leaves to
              search, set at query time allows user to tune search performance.
              This value increase result in both search accuracy and latency
              increase. The value should be between 0.0 and 1.0.

        Returns:
            "SearchNearestEntitiesResponse" - SearchNearestEntitiesResponse
            object. It is intermediate class that can be further converted by
            to_dict() or to_proto()
        """
        self.wait()
        if entity_id:
            embedding = None
        elif embedding_value:
            embedding = fos_service.NearestNeighborQuery.Embedding(
                value=embedding_value
            )
        else:
            raise ValueError(
                "Either entity_id or embedding_value needs to be provided for search."
            )
        response = self._get_online_store_client().search_nearest_entities(
            request=fos_service.SearchNearestEntitiesRequest(
                feature_view=self.resource_name,
                query=fos_service.NearestNeighborQuery(
                    entity_id=entity_id,
                    embedding=embedding,
                    neighbor_count=neighbor_count,
                    string_filters=string_filters,
                    per_crowding_attribute_neighbor_count=per_crowding_attribute_neighbor_count,  # pylint: disable=line-too-long
                    parameters=fos_service.NearestNeighborQuery.Parameters(
                        approximate_neighbor_candidates=approximate_neighbor_candidates,
                        leaf_nodes_search_fraction=leaf_nodes_search_fraction,
                    ),
                ),
                return_full_entity=return_full_entity,
            ),
            timeout=request_timeout,
        )
        return fs_utils.SearchNearestEntitiesResponse(response)

    class FeatureViewSync(base.VertexAiResourceNounWithFutureManager):
        """Class for managing Feature View Sync resources."""

        client_class = utils.FeatureOnlineStoreAdminClientWithOverride

        _resource_noun = "featureViewSyncs"
        _getter_method = "get_feature_view_sync"
        _list_method = "list_feature_view_syncs"
        _delete_method = "delete_feature_view"
        _sync_method = "sync_feature_view"
        _parse_resource_name_method = "parse_feature_view_sync_path"
        _format_resource_name_method = "feature_view_sync_path"
        _gca_resource: gca_feature_view_sync.FeatureViewSync

        def __init__(
            self,
            name: str,
            project: Optional[str] = None,
            location: Optional[str] = None,
            credentials: Optional[auth_credentials.Credentials] = None,
        ):
            """Retrieves an existing managed feature view sync.

            Args:
                name: The resource name
                  (`projects/.../locations/.../featureOnlineStores/.../featureViews/.../featureViewSyncs/...`)
                project: Project to retrieve the feature view from. If unset, the
                  project set in aiplatform.init will be used.
                location: Location to retrieve the feature view from. If not set,
                  location set in aiplatform.init will be used.
                credentials: Custom credentials to use to retrieve this feature view.
                  Overrides credentials set in aiplatform.init.
            """
            super().__init__(
                project=project,
                location=location,
                credentials=credentials,
                resource_name=name,
            )

            if not re.fullmatch(
                r"projects/.+/locations/.+/featureOnlineStores/.+/featureViews/.+/featureViewSyncs/.+",
                name,
            ):
                raise ValueError(
                    "name need to specify the fully qualified"
                    + " feature_view_sync resource path."
                )

            self._gca_resource = getattr(self.api_client, self._getter_method)(
                name=name, retry=base._DEFAULT_RETRY
            )

        @classmethod
        def sync_method(cls) -> str:
            """Returns the sync method."""
            return cls._sync_method



# Copyright 2016 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapper for streaming results."""

from google.cloud import exceptions
from google.protobuf.struct_pb2 import ListValue
from google.protobuf.struct_pb2 import Value

from google.cloud.spanner_v1 import PartialResultSet
from google.cloud.spanner_v1 import ResultSetMetadata
from google.cloud.spanner_v1 import TypeCode
from google.cloud.spanner_v1._helpers import _get_type_decoder, _parse_nullable


class StreamedResultSet(object):
    """Process a sequence of partial result sets into a single set of row data.

    :type response_iterator:
    :param response_iterator:
        Iterator yielding
        :class:`~google.cloud.spanner_v1.types.PartialResultSet`
        instances.

    :type source: :class:`~google.cloud.spanner_v1.snapshot.Snapshot`
    :param source: Deprecated. Snapshot from which the result set was fetched.
    """

    def __init__(
        self,
        response_iterator,
        source=None,
        column_info=None,
        lazy_decode: bool = False,
    ):
        self._response_iterator = response_iterator
        self._rows = []  # Fully-processed rows
        self._metadata = None  # Until set from first PRS
        self._stats = None  # Until set from last PRS
        self._current_row = []  # Accumulated values for incomplete row
        self._pending_chunk = None  # Incomplete value
        self._column_info = column_info  # Column information
        self._field_decoders = None
        self._lazy_decode = lazy_decode  # Return protobuf values
        self._done = False

    @property
    def fields(self):
        """Field descriptors for result set columns.

        :rtype: list of :class:`~google.cloud.spanner_v1.types.StructType.Field`
        :returns: list of fields describing column names / types.
        """
        return self._metadata.row_type.fields

    @property
    def metadata(self):
        """Result set metadata

        :rtype: :class:`~google.cloud.spanner_v1.types.ResultSetMetadata`
        :returns: structure describing the results
        """
        if self._metadata:
            return ResultSetMetadata.wrap(self._metadata)
        return None

    @property
    def stats(self):
        """Result set statistics

        :rtype:
           :class:`~google.cloud.spanner_v1.types.ResultSetStats`
        :returns: structure describing status about the response
        """
        return self._stats

    @property
    def _decoders(self):
        if self._field_decoders is None:
            if self._metadata is None:
                raise ValueError("iterator not started")
            self._field_decoders = [
                _get_type_decoder(field.type_, field.name, self._column_info)
                for field in self.fields
            ]
        return self._field_decoders

    def _merge_chunk(self, value):
        """Merge pending chunk with next value.

        :type value: :class:`~google.protobuf.struct_pb2.Value`
        :param value: continuation of chunked value from previous
                      partial result set.

        :rtype: :class:`~google.protobuf.struct_pb2.Value`
        :returns: the merged value
        """
        current_column = len(self._current_row)
        field = self.fields[current_column]
        merged = _merge_by_type(self._pending_chunk, value, field.type_)
        self._pending_chunk = None
        return merged

    def _merge_values(self, values):
        """Merge values into rows.

        :type values: list of :class:`~google.protobuf.struct_pb2.Value`
        :param values: non-chunked values from partial result set.
        """
        decoders = self._decoders
        width = len(self.fields)
        index = len(self._current_row)
        for value in values:
            if self._lazy_decode:
                self._current_row.append(value)
            else:
                self._current_row.append(_parse_nullable(value, decoders[index]))
            index += 1
            if index == width:
                self._rows.append(self._current_row)
                self._current_row = []
                index = 0

    def _consume_next(self):
        """Consume the next partial result set from the stream.

        Parse the result set into new/existing rows in :attr:`_rows`
        """
        response = next(self._response_iterator)
        response_pb = PartialResultSet.pb(response)

        if self._metadata is None:  # first response
            self._metadata = response_pb.metadata

        if response_pb.HasField("stats"):  # last response
            self._stats = response.stats

        values = list(response_pb.values)
        if self._pending_chunk is not None:
            values[0] = self._merge_chunk(values[0])

        if response_pb.chunked_value:
            self._pending_chunk = values.pop()

        self._merge_values(values)

        if response_pb.last:
            self._done = True

    def __iter__(self):
        while True:
            iter_rows, self._rows[:] = self._rows[:], ()
            while iter_rows:
                yield iter_rows.pop(0)
            if self._done:
                return
            try:
                self._consume_next()
            except StopIteration:
                return

    def decode_row(self, row: []) -> []:
        """Decodes a row from protobuf values to Python objects. This function
           should only be called for result sets that use ``lazy_decoding=True``.
           The array that is returned by this function is the same as the array
           that would have been returned by the rows iterator if ``lazy_decoding=False``.

        :returns: an array containing the decoded values of all the columns in the given row
        """
        if not hasattr(row, "__len__"):
            raise TypeError("row", "row must be an array of protobuf values")
        decoders = self._decoders
        return [
            _parse_nullable(row[index], decoders[index]) for index in range(len(row))
        ]

    def decode_column(self, row: [], column_index: int):
        """Decodes a column from a protobuf value to a Python object. This function
           should only be called for result sets that use ``lazy_decoding=True``.
           The object that is returned by this function is the same as the object
           that would have been returned by the rows iterator if ``lazy_decoding=False``.

        :returns: the decoded column value
        """
        if not hasattr(row, "__len__"):
            raise TypeError("row", "row must be an array of protobuf values")
        decoders = self._decoders
        return _parse_nullable(row[column_index], decoders[column_index])

    def one(self):
        """Return exactly one result, or raise an exception.

        :raises: :exc:`NotFound`: If there are no results.
        :raises: :exc:`ValueError`: If there are multiple results.
        :raises: :exc:`RuntimeError`: If consumption has already occurred,
            in whole or in part.
        """
        answer = self.one_or_none()
        if answer is None:
            raise exceptions.NotFound("No rows matched the given query.")
        return answer

    def one_or_none(self):
        """Return exactly one result, or None if there are no results.

        :raises: :exc:`ValueError`: If there are multiple results.
        :raises: :exc:`RuntimeError`: If consumption has already occurred,
            in whole or in part.
        """
        # Sanity check: Has consumption of this query already started?
        # If it has, then this is an exception.
        if self._metadata is not None:
            raise RuntimeError(
                "Can not call `.one` or `.one_or_none` after "
                "stream consumption has already started."
            )

        # Consume the first result of the stream.
        # If there is no first result, then return None.
        iterator = iter(self)
        try:
            answer = next(iterator)
        except StopIteration:
            return None

        # Attempt to consume more. This should no-op; if we get additional
        # rows, then this is an error case.
        try:
            next(iterator)
            raise ValueError("Expected one result; got more.")
        except StopIteration:
            return answer

    def to_dict_list(self):
        """Return the result of a query as a list of dictionaries.
        In each dictionary the key is the column name and the value is the
        value of the that column in a given row.

        :rtype:
           :class:`list of dict`
        :returns: result rows as a list of dictionaries
        """
        rows = []
        for row in self:
            rows.append(
                {
                    column: value
                    for column, value in zip(
                        [column.name for column in self._metadata.row_type.fields], row
                    )
                }
            )
        return rows


class Unmergeable(ValueError):
    """Unable to merge two values.

    :type lhs: :class:`~google.protobuf.struct_pb2.Value`
    :param lhs: pending value to be merged

    :type rhs: :class:`~google.protobuf.struct_pb2.Value`
    :param rhs: remaining value to be merged

    :type type_: :class:`~google.cloud.spanner_v1.types.Type`
    :param type_: field type of values being merged
    """

    def __init__(self, lhs, rhs, type_):
        message = "Cannot merge %s values: %s %s" % (
            TypeCode(type_.code),
            lhs,
            rhs,
        )
        super(Unmergeable, self).__init__(message)


def _unmergeable(lhs, rhs, type_):
    """Helper for '_merge_by_type'."""
    raise Unmergeable(lhs, rhs, type_)


def _merge_float64(lhs, rhs, type_):
    """Helper for '_merge_by_type'."""
    lhs_kind = lhs.WhichOneof("kind")
    if lhs_kind == "string_value":
        return Value(string_value=lhs.string_value + rhs.string_value)
    rhs_kind = rhs.WhichOneof("kind")
    array_continuation = (
        lhs_kind == "number_value"
        and rhs_kind == "string_value"
        and rhs.string_value == ""
    )
    if array_continuation:
        return lhs
    raise Unmergeable(lhs, rhs, type_)


def _merge_string(lhs, rhs, type_):
    """Helper for '_merge_by_type'."""
    return Value(string_value=lhs.string_value + rhs.string_value)


_UNMERGEABLE_TYPES = (TypeCode.BOOL,)


def _merge_array(lhs, rhs, type_):
    """Helper for '_merge_by_type'."""
    element_type = type_.array_element_type
    if element_type.code in _UNMERGEABLE_TYPES:
        # Individual values cannot be merged, just concatenate
        lhs.list_value.values.extend(rhs.list_value.values)
        return lhs
    lhs, rhs = list(lhs.list_value.values), list(rhs.list_value.values)

    # Sanity check: If either list is empty, short-circuit.
    # This is effectively a no-op.
    if not len(lhs) or not len(rhs):
        return Value(list_value=ListValue(values=(lhs + rhs)))

    first = rhs.pop(0)
    if first.HasField("null_value"):  # can't merge
        lhs.append(first)
    else:
        last = lhs.pop()
        if last.HasField("null_value"):
            lhs.append(last)
            lhs.append(first)
        else:
            try:
                merged = _merge_by_type(last, first, element_type)
            except Unmergeable:
                lhs.append(last)
                lhs.append(first)
            else:
                lhs.append(merged)
    return Value(list_value=ListValue(values=(lhs + rhs)))


def _merge_struct(lhs, rhs, type_):
    """Helper for '_merge_by_type'."""
    fields = type_.struct_type.fields
    lhs, rhs = list(lhs.list_value.values), list(rhs.list_value.values)

    # Sanity check: If either list is empty, short-circuit.
    # This is effectively a no-op.
    if not len(lhs) or not len(rhs):
        return Value(list_value=ListValue(values=(lhs + rhs)))

    candidate_type = fields[len(lhs) - 1].type_
    first = rhs.pop(0)
    if first.HasField("null_value") or candidate_type.code in _UNMERGEABLE_TYPES:
        lhs.append(first)
    else:
        last = lhs.pop()
        if last.HasField("null_value"):
            lhs.append(last)
            lhs.append(first)
        else:
            try:
                merged = _merge_by_type(last, first, candidate_type)
            except Unmergeable:
                lhs.append(last)
                lhs.append(first)
            else:
                lhs.append(merged)
    return Value(list_value=ListValue(values=lhs + rhs))


_MERGE_BY_TYPE = {
    TypeCode.ARRAY: _merge_array,
    TypeCode.BOOL: _unmergeable,
    TypeCode.BYTES: _merge_string,
    TypeCode.DATE: _merge_string,
    TypeCode.FLOAT64: _merge_float64,
    TypeCode.FLOAT32: _merge_float64,
    TypeCode.INT64: _merge_string,
    TypeCode.STRING: _merge_string,
    TypeCode.STRUCT: _merge_struct,
    TypeCode.TIMESTAMP: _merge_string,
    TypeCode.NUMERIC: _merge_string,
    TypeCode.JSON: _merge_string,
    TypeCode.PROTO: _merge_string,
    TypeCode.INTERVAL: _merge_string,
    TypeCode.ENUM: _merge_string,
}


def _merge_by_type(lhs, rhs, type_):
    """Helper for '_merge_chunk'."""
    merger = _MERGE_BY_TYPE[type_.code]
    return merger(lhs, rhs, type_)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# TODO: use opentelemetry-semantic-conventions package for these constants once it has
# stabilized. Right now, pinning an unstable version would cause dependency conflicts for
# users so these are copied in.
class ResourceAttributes:
    AWS_EC2 = "aws_ec2"
    CLOUD_ACCOUNT_ID = "cloud.account.id"
    CLOUD_AVAILABILITY_ZONE = "cloud.availability_zone"
    CLOUD_PLATFORM_KEY = "cloud.platform"
    CLOUD_PROVIDER = "cloud.provider"
    CLOUD_REGION = "cloud.region"
    FAAS_INSTANCE = "faas.instance"
    FAAS_NAME = "faas.name"
    FAAS_VERSION = "faas.version"
    GCP_APP_ENGINE = "gcp_app_engine"
    GCP_CLOUD_FUNCTIONS = "gcp_cloud_functions"
    GCP_CLOUD_RUN = "gcp_cloud_run"
    GCP_COMPUTE_ENGINE = "gcp_compute_engine"
    GCP_KUBERNETES_ENGINE = "gcp_kubernetes_engine"
    HOST_ID = "host.id"
    HOST_NAME = "host.name"
    HOST_TYPE = "host.type"
    K8S_CLUSTER_NAME = "k8s.cluster.name"
    K8S_CONTAINER_NAME = "k8s.container.name"
    K8S_NAMESPACE_NAME = "k8s.namespace.name"
    K8S_NODE_NAME = "k8s.node.name"
    K8S_POD_NAME = "k8s.pod.name"
    SERVICE_INSTANCE_ID = "service.instance.id"
    SERVICE_NAME = "service.name"
    SERVICE_NAMESPACE = "service.namespace"


AWS_ACCOUNT = "aws_account"
AWS_EC2_INSTANCE = "aws_ec2_instance"
CLUSTER_NAME = "cluster_name"
CONTAINER_NAME = "container_name"
GCE_INSTANCE = "gce_instance"
GENERIC_NODE = "generic_node"
GENERIC_TASK = "generic_task"
INSTANCE_ID = "instance_id"
JOB = "job"
K8S_CLUSTER = "k8s_cluster"
K8S_CONTAINER = "k8s_container"
K8S_NODE = "k8s_node"
K8S_POD = "k8s_pod"
LOCATION = "location"
NAMESPACE = "namespace"
NAMESPACE_NAME = "namespace_name"
NODE_ID = "node_id"
NODE_NAME = "node_name"
POD_NAME = "pod_name"
REGION = "region"
TASK_ID = "task_id"
ZONE = "zone"
UNKNOWN_SERVICE_PREFIX = "unknown_service"
